#include "tla2022.h"
#include "dfbc.h"
#include "CH58x_gpio.h"
#include "comm.h"
#include "hw.h"

enum FullScaleRange currentFSR = FSR_2_048V;
enum OperatingMode currentMode_ ;
uint16_t tla202xConf;
uint8_t channelToRead = 0;

uint16_t tla202xReadReg(uint8_t reg){
    int ret;
    uint8_t data[2];
    uint16_t value = 0;
    ret = i2c_write_to(TLA202X_I2C_ADDR ,&reg,1, true, true);
    ret = i2c_read_from(TLA202X_I2C_ADDR, data, 2, true, 1000);
    if (ret == 2){
        value = data[0]<<8 | data[1];
        //printf("reg[%d]=%04X\n",reg,value);
    }else{
        //DBG("read 0x%x fail!!\n",reg);
    }

    return value;
}

int tla202xWriteReg(uint8_t reg,uint16_t value){
    uint8_t data[3]={0,0,0};
    int ret;
    data[0] = reg;

    data[2] =  value|0xff;
    data[1] =  value>>8;
    ret = i2c_write_to(TLA202X_I2C_ADDR, data, 3, true, true);
    if(ret != 0){
        DBG("writeReg 0x%x fail!!\n",reg);
    }
    return ret;
}

void tla202xBegin(){
	int ret;
    GPIOPinRemap(ENABLE, RB_PIN_I2C);
    DelayMs(1);
    i2c_app_init(0);
    ret = tla202xWriteReg(TLA202X_CONF_REG, 0x85FF);
    tla202xConf = tla202xReadReg(TLA202X_CONF_REG);

    tla202xSetFSR(FSR_2_048V);
    tla202xSetDR(DR_128SPS);
    tla202xSetOperatingMode(OP_SINGLE);
}

void tla202xSetFSR(enum FullScaleRange range) {
    currentFSR = range;
    tla202xConf &= ~0x0E00;
    tla202xConf |= range << 9;
    tla202xWriteReg(TLA202X_CONF_REG, tla202xConf);
}

void tla202xSetDR(enum DataRate rate) {
	tla202xConf |= rate << 5;
	tla202xWriteReg(TLA202X_CONF_REG, tla202xConf);
}

int16_t tla202xADC0() {
    // this only needs to run when in single shot.
	int times = 50;
    if (currentMode_ == OP_SINGLE) {
    	int ret;
        // write 1 to OS bit to start conv
    	tla202xConf |= 0x8000;
    	ret = tla202xWriteReg(TLA202X_CONF_REG, tla202xConf);
    	if(ret) return 0;
        // OS bit will be 0 until conv is done.
    	while (((tla202xReadReg(TLA202X_CONF_REG) & 0x8000) == 0)) {
            DelayMs(1);
            if(--times == 0) return 0;
        }
    }

    uint16_t in_data = tla202xReadReg(TLA202X_DATA_REG);
    in_data >>= 4;

    if (in_data & (1 << 11)) {
        in_data |= 0xF000;
    } else {
        in_data &= ~0xF000;
    }
    if(in_data > 3000){
    	in_data = 0;
    }
    // now store it as a signed 16 bit int.
    int16_t ret = in_data;

    // default Full Scale Range is -2.048V to 2.047V.
    // our 12bit 2's complement goes from -2048 to 2047
    return ret;
}
void tla202xSetOperatingMode(enum OperatingMode mode) {

    currentMode_ = mode;
    // clear MODE bit (8) (continous conv)
    tla202xConf &= ~(1 << 8);
    if (mode == OP_SINGLE) tla202xConf |= (1 << 8);
    tla202xWriteReg(TLA202X_CONF_REG, tla202xConf);
}
void tla202xSetMuxConfig(enum MultiplexerConfig option, bool write) {
    // clear MUX bits
	tla202xConf &= ~0x7000;
	tla202xConf |= option << 12;
	tla202xWriteReg(TLA202X_CONF_REG, tla202xConf);
}

int16_t analogRead(uint8_t channel) {
    //Writes new settings immediately only if in continous mode.
    //In single mode analogRead writes config.
   // enum MultiplexerConfig muxCfg = (enum MultiplexerConfig)(channel - 4);
    //setMuxConfig(muxCfg, currentMode_ == OP_CONTINUOUS);
    return tla202xADC0();
}
int16_t read_adc() {
    uint16_t in_data = tla202xReadReg(TLA202X_DATA_REG);

    int16_t ret = (int16_t) in_data;
    ret>>=4;

    return ret;
}
float getCurrentFullRangeVoltage() {
    uint16_t shifted = 8192 >> currentFSR;

    // Special case
    if (currentFSR == 0) {
        shifted = 6144;
    }

    return (float)shifted / 1000.0f;
}
float getVoltageResolution() {
    switch (currentFSR) {
        case FSR_6_144V: return 0.003f;
        case FSR_4_096V: return 0.002f;
        case FSR_2_048V: return 0.001f;
        case FSR_1_024V: return 0.0005f;
        case FSR_0_512V: return 0.00025f;
        case FSR_0_256V: return 0.000125f;
    }
    return 0.0f;
}

float voltageRead(uint8_t channel) {
    if (channel > 3) return 0.0f;
    int16_t val = analogRead(channel);

    float fsrV = getCurrentFullRangeVoltage() - getVoltageResolution();

    //return (val - 0) * fsrV / (2047l - 0);
    return (float)val * fsrV / 2047.0f; //map

    //long converted = map((long)val, -2048, 2047, 0, fsrV * 100000);
    //long converted = map((long)val, 0l, 2047l, 0l, fsrV * 100000);
    // return (float)converted / 100000.0f;
}
