#include "NT3H.h"

static uint8_t sendBuf[NT3H_I2C_BUFSIZE] = {0};

#define READ  1
#define WRITE 0

/***************************************************
 *                 Initialization
 ***************************************************/

int NT3H_init(void)
{
    // 初始化中断引脚
    GPIOB_ModeCfg(NT3H_INT_PIN, GPIO_ModeIN_PU);
    GPIOB_ITModeCfg(NT3H_INT_PIN, GPIO_ITMode_FallEdge);
    PFIC_EnableIRQ(GPIO_B_IRQn);
    // 初始化电源引脚
    GPIOB_ModeCfg(NT3H_PWR_PIN, GPIO_ModeOut_PP_20mA);
    return NT3H_SUCCESS;
}

int NT3H_check_id(uint8_t *pBuf)
{
    // 检查设备制造商
    uint8_t devInfo[7] = {0};
    if (NT3H_SUCCESS != NT3H_read(NT3H_BLOCK_INFO, devInfo, 7))
    {
        return NT3H_ERR_I2C_TRANSMITTED_FAILURE;
    }
   // PRINT("\r\n%02X-%02X-%02X-%02X-%02X-%02X-%02X\r\n", devInfo[0], devInfo[1], devInfo[2], devInfo[3], devInfo[4],
     //     devInfo[5], devInfo[6]);
    if (0x04 == devInfo[0])
    {
        //PRINT("NXP ID Confirmed\r\n");
        if (pBuf != NULL)
            memcpy(pBuf, devInfo, NT3H_NXPID_LENGTH);
        return NT3H_SUCCESS;
    }
    else
    {
        PRINT("NXP ID Unconfirmed\r\n");
        return NT3H_ERR_ILLEGAL_NXP_ID;
    }
}

int NT3H_check_cc(uint8_t *pDat)
{
    // 检查 CC File
    uint8_t devInfo[16] = {0};
    if (NT3H_SUCCESS != NT3H_read(NT3H_BLOCK_INFO, devInfo, NT3H_BLOCK_SIZE))
    {
        return NT3H_ERR_I2C_TRANSMITTED_FAILURE;
    }
    if (0x04 != devInfo[0])
    {
        return NT3H_ERR_ILLEGAL_NXP_ID;
    }
    if (memcmp(devInfo + 12, pDat, 4))
    {
        devInfo[0] = (NT3H_I2C_ADDR << 1);
        memcpy(devInfo + 12, pDat, 4);
        if (NT3H_SUCCESS != NT3H_write(NT3H_BLOCK_INFO, devInfo, NT3H_BLOCK_SIZE))
        {
            return NT3H_ERR_I2C_TRANSMITTED_FAILURE;
        }
    }
    return NT3H_SUCCESS;
}

int NT3H_FD_config(uint8_t onEvt, uint8_t offEvt, uint8_t lastNDEFPage)
{
    const uint8_t mask = BIT(5) | BIT(4) | BIT(3) | BIT(2);
    uint8_t       configBlock[NT3H_BLOCK_SIZE];
    if (NT3H_SUCCESS != NT3H_read(NT3H_BLOCK_CONF, configBlock, NT3H_BLOCK_SIZE))
    {
        return NT3H_ERR_I2C_TRANSMITTED_FAILURE;
    }

    // 先读取配置区设置，如果配置未变，返回
    if ((configBlock[NT3H_REG_NC_REG] & mask) == (onEvt | offEvt))
    {
        if (NT3H_FD_LASTPAGE_READED == offEvt)
        {
            if (configBlock[NT3H_REG_LAST_NDEF_BLOCK] == lastNDEFPage)
            {
               // PRINT("DF NOT CHANGE!\r\n");
                return NT3H_SUCCESS;
            }
        }
        else
        {
           // PRINT("DF NOT CHANGE!\r\n");
            return NT3H_SUCCESS;
        }
    }

    // 否则修改配置区
    configBlock[NT3H_REG_NC_REG] &= (~mask);
    configBlock[NT3H_REG_NC_REG] |= (onEvt | offEvt | 1); // (|1) 避免锁住 NFC 写权限
    if (NT3H_FD_LASTPAGE_READED == offEvt)
    {
        configBlock[NT3H_REG_LAST_NDEF_BLOCK] = lastNDEFPage;
    }
    configBlock[NT3H_REG_REGNS_LOCK] = 0; // 避免误锁
    if (NT3H_SUCCESS != NT3H_write(NT3H_BLOCK_CONF, configBlock, NT3H_BLOCK_SIZE))
    {
        return NT3H_ERR_I2C_TRANSMITTED_FAILURE;
    }

    // 同步配置到会话寄存器
    if (NT3H_SUCCESS != NT3H_write_reg(NT3H_REG_NC_REG, mask, offEvt | onEvt))
    {
        return NT3H_ERR_I2C_TRANSMITTED_FAILURE;
    }
    if (NT3H_FD_LASTPAGE_READED == offEvt)
    {
        NT3H_write_reg(NT3H_REG_LAST_NDEF_BLOCK, 0xFF, lastNDEFPage);
    }
    return NT3H_SUCCESS;
}

int NT3H_auth(uint8_t auth, uint8_t access, uint8_t *pwd, uint8_t *pack, uint8_t pt)
{
    uint8_t defaultBlockLock[NT3H_BLOCK_SIZE] = NT3H_DEFAULT_BLOCK_LOCK;
    defaultBlockLock[NT3H_AUTH_AUTH]          = auth;

    uint8_t defaultBlockPWD[16]       = NT3H_DEFAULT_BLOCK_PWD;
    defaultBlockPWD[NT3H_AUTH_ACCESS] = access;
    memcpy(defaultBlockPWD + NT3H_AUTH_PWD, pwd, NT3H_PWD_LENGTH);
    memcpy(defaultBlockPWD + NT3H_AUTH_PACK, pack, NT3H_PACK_LENGTH);
    defaultBlockPWD[NT3H_AUTH_PT_I2C] = pt;

    if (NT3H_SUCCESS != NT3H_write(NT3H_BLOCK_PWD, defaultBlockPWD, NT3H_BLOCK_SIZE))
    {
        return NT3H_ERR_I2C_TRANSMITTED_FAILURE;
    }

    if (NT3H_SUCCESS != NT3H_write(NT3H_BLOCK_LOCK, defaultBlockLock, NT3H_BLOCK_SIZE))
    {
        return NT3H_ERR_I2C_TRANSMITTED_FAILURE;
    }
    return NT3H_SUCCESS;
}

/***************************************************
 *                 Check NTag Status
 ***************************************************/

int NT3H_is_E2PROM_available(void)
{
    uint8_t regValue;
    if (NT3H_SUCCESS != NT3H_read_reg(NT3H_REG_REGNS_LOCK, &regValue))
    {
        return false;
    }
    return regValue & (BIT(1));
}

int NT3H_is_radio_field_exist(void)
{
    uint8_t regValue;
    if (NT3H_SUCCESS != NT3H_read_reg(NT3H_REG_REGNS_LOCK, &regValue))
    {
        return false;
    }
    return regValue & (BIT(0));
}

int NT3H_is_RF_locked(void)
{
    uint8_t regValue;
    if (NT3H_SUCCESS != NT3H_read_reg(NT3H_REG_REGNS_LOCK, &regValue))
    {
        return false;
    }
    return regValue & BIT(5);
}

int NT3H_is_write_ready(void)
{
    uint8_t regValue;
    if (NT3H_SUCCESS != NT3H_read_reg(NT3H_REG_REGNS_LOCK, &regValue))
    {
        return false;
    }
    if ((regValue & BIT(5)) || (regValue & BIT(1)))
    {
        return false;
    }
    if (NT3H_SUCCESS != NT3H_read_reg(NT3H_REG_REGNS_LOCK, &regValue))
    {
        return false;
    }
    if (regValue & BIT(1)) // 直通模式不能访问 E2PROM
    {
        return false;
    }
    return true;
}

/***************************************************
 *                       NDEF
 ***************************************************/

int NT3H_write_NDEF(uint8_t memAddr, uint8_t *pDat, size_t len)
{
    int ret, offset;
    // 初始化一个 NDEF 消息，并生成一个 TLV
    sNDEFTLV_t msg;
    if (NDEF_SUCCESS != NDEF_Text_Init(&msg, pDat, len))
    {
        return NT3H_ERR_NDEF_MSG_INIT_FAILURE;
    }
    //PRINT("Len: %d", msg.TLVLen);
    if (msg.TLVLen > 870)
    {
        return NT3H_ERR_BUF_OVERFLOW;
    }
    // 为 NDEF 添加 TLV Term
    msg.TLVBuf[msg.TLVLen] = TLVTerm[0];
    // 写入 E2PROM
    NT3H_write(memAddr, msg.TLVBuf, msg.TLVLen + 1);
}

int NT3H_read_NDEF(uint8_t memAddr, uint8_t *pBuf, size_t len)
{
    // ---Read a Tag---
    sNDEFTLV_t msg;
    NT3H_read(NT3H_BLOCK_USER0, msg.TLVBuf, 4);
    if (msg.TLVBuf[0] == 03)
    {
        if (msg.TLVBuf[1] == 0xFF)
        {
            msg.TLVLen = ((uint16_t)msg.TLVBuf[2] << 8) + msg.TLVBuf[3] + 4;
        }
        else
        {
            msg.TLVLen = msg.TLVBuf[1] + 2;
        }
    }
    else
    {
        return NT3H_ERR_NDEF_UNDETECTED;
    }
    if (NT3H_SUCCESS != NT3H_read(NT3H_BLOCK_USER0, msg.TLVBuf, msg.TLVLen))
    {
        return NT3H_ERR_I2C_TRANSMITTED_FAILURE;
    }
    if (NDEF_SUCCESS != NDEF_Parsing(&msg))
    {
        return NT3H_ERR_NDEF_PARSING_FAILURE;
    }
    if (len < msg.record.payloadLength)
    {
        return NT3H_ERR_BUF_OVERFLOW;
    }
    memcpy(pBuf, msg.record.payload, msg.record.payloadLength);
    return msg.record.payloadLength;
}

/***************************************************
 *                    Pass-through
 ***************************************************/

int NT3H_open_pthru(uint8_t dir)
{
    uint8_t mask  = BIT(6) | BIT(5) | BIT(4) | BIT(3) | BIT(2) | BIT(0);
    uint8_t value = BIT(6) | NT3H_FD_PT_DATA_READY | NT3H_FD_PT_READ_COMPLET | dir;
    return NT3H_write_reg(NT3H_REG_NC_REG, mask, value);
}

int NT3H_close_pthru(void)
{
    uint8_t mask  = BIT(6) | BIT(5) | BIT(4) | BIT(3) | BIT(2) | BIT(0);
    uint8_t value = NT3H_FD_COMMUNICATION_START | NT3H_FD_FIELD_CLOSE;
    return NT3H_write_reg(NT3H_REG_NC_REG, mask, value);
}

int NT3H_is_pthru_write_ready(void)
{
    return !NT3H_is_RF_locked();
}

int NT3H_is_pthru_read_ready(void)
{
    uint8_t regValue;
    if (NT3H_SUCCESS != NT3H_read_reg(NT3H_REG_REGNS_LOCK, &regValue))
    {
        return false;
    }
    return regValue & BIT(4);
}

int NT3H_pthru_read(uint8_t *pBuf)
{
    return NT3H_read(NT3H_BLOCK_SRAM, pBuf, 64);
}

int NT3H_pthru_write(uint8_t *pDat)
{
    return NT3H_write(NT3H_BLOCK_SRAM, pDat, 64);
}

/***************************************************
 *                       E2PROM
 ***************************************************/

int NT3H_write(uint8_t memAddr, uint8_t *pDat, size_t len)
{
    int ret = 0;

    while (len > 0)
    {
        if (len >= NT3H_BLOCK_SIZE)
        {
            ret = NT3H_write_block(memAddr++, pDat, NT3H_BLOCK_SIZE);
        }
        else
        {
            ret = NT3H_write_block(memAddr++, pDat, len);
            return ret;
        }
        pDat += NT3H_BLOCK_SIZE;
        len -= NT3H_BLOCK_SIZE;
        if (NT3H_SUCCESS != ret)
            return ret;
    }
    return NT3H_SUCCESS;
}

int NT3H_read(uint8_t memAddr, uint8_t *pBuf, size_t len)
{
    int ret;

    while (len > 0)
    {
        if (len >= NT3H_BLOCK_SIZE)
        {
            ret = NT3H_read_block(memAddr++, pBuf, NT3H_BLOCK_SIZE);
        }
        else
        {
            ret = NT3H_read_block(memAddr++, pBuf, len);
            return ret;
        }
        pBuf += NT3H_BLOCK_SIZE;
        len -= NT3H_BLOCK_SIZE;
        if (NT3H_SUCCESS != ret)
            return ret;
    }
    return NT3H_SUCCESS;
}

int NT3H_write_block(uint8_t memAddr, uint8_t *pDat, size_t len)
{
    if ((memAddr > NT3H_BLOCK_CONF) && (memAddr < NT3H_BLOCK_USER1))
    {
        return NT3H_ERR_ILLEGAL_MEMADRR;
    }
    else if ((memAddr > NT3H_BLOCK_USER1_TERM) && (memAddr < NT3H_BLOCK_SRAM))
    {
        return NT3H_ERR_ILLEGAL_MEMADRR;
    }
    else if ((memAddr > NT3H_BLOCK_SRAM_TERM))
    {
        return NT3H_ERR_ILLEGAL_MEMADRR;
    } // 检查 memAddr 是否在合法范围内
    if (len > NT3H_BLOCK_SIZE)
        len = NT3H_BLOCK_SIZE;
   // PRINT("NT3H_is_write_ready!\r\n");
    uint8_t timeout = 100;
    while (true != NT3H_is_write_ready())
    {
        timeout--;
        DelayMs(1);
        if(timeout == 0 ){
            break;
        }
    } // 打断 E2PROM 写入可能导致损坏
    memset(sendBuf, 0, NT3H_I2C_BUFSIZE);
    sendBuf[0] = memAddr;
    memcpy(sendBuf + 1, pDat, len);
    if (0 != i2c_write_to(NT3H_I2C_ADDR, sendBuf, NT3H_I2C_BUFSIZE, TRUE, TRUE))
    {
        PRINT("I2C WRITE RAW FAILED\n");
        return NT3H_ERR_I2C_TRANSMITTED_FAILURE;
    }
#ifdef PRINT_COMM
    PRINT("Write Block: %d\r\n", memAddr);
    printBlock(sendBuf + 1, NT3H_BLOCK_SIZE);
#else
    DelayMs(4);
#endif
    return NT3H_SUCCESS;
}

int NT3H_read_block(uint8_t memAddr, uint8_t *pBuf, size_t len)
{
    if (len > NT3H_BLOCK_SIZE)
        len = NT3H_BLOCK_SIZE;
    int ret;
    ret = i2c_write_to(NT3H_I2C_ADDR, &memAddr, 1, TRUE, TRUE);
    if (0 != ret)
    {
        PRINT("READ NO ACK!\r\n");
        return NT3H_ERR_I2C_TRANSMITTED_FAILURE;
    }
    ret = i2c_read_from(NT3H_I2C_ADDR, (uint8_t *)pBuf, len, TRUE, 1000);
    if (0 < ret)
    {
#ifdef PRINT_COMM
        PRINT("Read Block: %d\r\n", memAddr);
        printBlock(pBuf, len);
#endif
        return NT3H_SUCCESS;
    }
    else
    {
        PRINT("I2C READ ERR: %d", ret);
        return NT3H_ERR_I2C_TRANSMITTED_FAILURE;
    }
}

int NT3H_write_reg(uint8_t regAddr, uint8_t regMask, uint8_t regData)
{
    memset(sendBuf, 0, NT3H_I2C_BUFSIZE);
    sendBuf[0] = NT3H_BLOCK_SESS;
    sendBuf[1] = regAddr;
    sendBuf[2] = regMask;
    sendBuf[3] = regData;
    if (0 != i2c_write_to(NT3H_I2C_ADDR, sendBuf, 4, TRUE, TRUE))
    {
        PRINT("WRITE FAILED\n");
        return NT3H_ERR_I2C_TRANSMITTED_FAILURE;
    }
#ifdef PRINT_COMM
    PRINT("Write Reg %02X with %02X %02X\r\n", regAddr, regMask, regData);
#endif
    return NT3H_SUCCESS;
}

int NT3H_read_reg(uint8_t regAddr, uint8_t *pBuf)
{
    int ret = 0;
    memset(sendBuf, 0, NT3H_I2C_BUFSIZE);
    sendBuf[0] = NT3H_BLOCK_SESS;
    sendBuf[1] = regAddr;

    ret = i2c_write_to(NT3H_I2C_ADDR, sendBuf, 2, TRUE, TRUE);
    if (0 != ret)
    {
        return NT3H_ERR_I2C_TRANSMITTED_FAILURE;
    }
    ret = i2c_read_from(NT3H_I2C_ADDR, (uint8_t *)pBuf, 1, TRUE, 1000);
    if (0 < ret)
    {
#ifdef PRINT_COMM
        PRINT("Read Reg %02X get %02X\r\n", regAddr, *pBuf);
#endif
        return NT3H_SUCCESS;
    }
    else
    {
        return NT3H_ERR_I2C_TRANSMITTED_FAILURE;
    }
}

int NT3H_clear_blocks(uint8_t memAddr, uint8_t endMema)
{
    uint8_t blockClaer[NT3H_BLOCK_SIZE];
    memset(blockClaer, 0, sizeof(blockClaer));
    for (uint8_t i = memAddr; i < endMema + 1; i++)
    {
        if (NT3H_SUCCESS != NT3H_write_block(i, blockClaer, NT3H_BLOCK_SIZE))
        {
            return NT3H_ERR_I2C_TRANSMITTED_FAILURE;
        }
    }
    return NT3H_SUCCESS;
}

/***************************************************
 *                     Debug Tools
 ***************************************************/

void printMem(uint8_t beginAddr, uint8_t endAddr)
{
    uint8_t memAddr = beginAddr;
    uint8_t pBuf[16];
    while (memAddr < (endAddr + 1))
    {
        NT3H_read_block(memAddr++, pBuf, 16);
    }
}

void printBlock(uint8_t *pBuf, size_t len)
{
    uint8_t row = len / 4;
    if (len % 4 > 0)
        row++;
    PRINT("+-------------+\r\n");
    for (uint8_t i = 0; i < row; i++)
    {
        PRINT("| ");
        for (uint8_t j = 0; j < 4; j++)
        {
            uint8_t index = 4 * i + j;
            if (index < len)
                PRINT("%02X ", pBuf[index]);
            else
                PRINT("   ");
        }
        PRINT("|\r\n");
    }
    PRINT("+-------------+\r\n");
}
