#ifndef __DFBC_UTILS_H__
#define __DFBC_UTILS_H__

#include <stdint.h>
#include <stdlib.h>

/**************************************************
 *                     DEBUG
 **************************************************/
//#define DEBUG_DFBC
#ifdef DEBUG_DFBC
#include <stdio.h>
extern const char *errors[];
#define DDBG(fmt, args...)     printf(fmt"\n", ##args)
#define DBG_NLN(fmt, args...) printf(fmt, ##args)
#define DBG_DTL(fmt, args...) printf("%s:%d: %s(): " fmt "\n", __FILE__, __LINE__, __FUNCTION__, ##args)
#define DBG_HEX(array, size)                                                                                           \
    do {                                                                                                               \
        for (int i_ = 0; i_ < size; i_++) {                                                                            \
            printf("%02x ", (array)[i_]);                                                                              \
        }                                                                                                              \
        printf("\n");                                                                                                  \
    } while (0)
#define DBG_BRK()    while (0)
#define DBG_ERR(err) printf("%s\n", errors[err]);
#define ERR_STR(err) #err
#define TAB          "    "
#define TAB2         "    ""    "
#define TAB3         "    ""    ""    "
#define EDL          "\n"


#else
#define DDBG(fmt, args...)
#define DBG_NLN(fmt, args...)   // print debug info without '\n'
#define DBG_DTL(fmt, args...)   // detail(file, line, function)
#define DBG_HEX(array, size)    // print uint8_t array in hex mode
#define DBG_BRK()               // a util to help set up break point
#define DBG_ASM()
#define ERR_STR(err)
#define TAB
#define TAB2
#define TAB3
#define EDL
#endif   // DEBUG

/**************************************************
 *                     STRING
 **************************************************/

typedef struct sCTypeString {
    char  *data;
    size_t size;
} cstr_t;

size_t tocstr(const char *str, cstr_t *cstr);

/**************************************************
 *                     BUFFER
 **************************************************/

typedef struct {
    uint8_t *head;
    uint8_t *curr;
    size_t   size;
} buff_t;

#define U8A2U32(x)                                                                                                     \
    ((uint32_t) * ((x)) | (uint32_t) * ((x) + 1) << 8 | (uint32_t) * ((x) + 2) << 16 | (uint32_t) * ((x) + 3) << 24)

#define U322U8A(x, idx)                   (uint8_t)((x) >> ((idx) * 8))

#define BUFF_RST(buff)                    ((buff).curr = (buff).head)
#define BUFF_MOVE(buff, offset)           ((buff).curr += offset)
#define BUFF_GET_HEAD(buff)               ((buff).head)
#define BUFF_GET_CURR(buff)               ((buff).curr)
#define BUFF_GET_SIZE(buff)               ((buff).size)
#define BUFF_SET_HEAD(buff, value)        ((buff).head = (uint8_t *)value)
#define BUFF_SET_CURR(buff, value)        ((buff).curr = (uint8_t *)value)
#define BUFF_SET_SIZE(buff, value)        ((buff).size = (size_t)value)
#define BUFF_OFFSET(buff)                 (BUFF_GET_CURR(buff) - BUFF_GET_HEAD(buff))
#define BUFF_SETV_MV(buff, value)         (*(BUFF_GET_CURR(buff)++) = (value))
#define BUFF_SETV_AT(buff, offset, value) do{(*(BUFF_GET_HEAD(buff) + offset) = (value)); DDBG("SET 0x%X to 0x%X", (BUFF_GET_HEAD(buff) + (offset)), (value));}while(0)
#define BUFF_COPY_MV(buff, src, len)                                                                                   \
    do {                                                                                                               \
        memcpy(BUFF_GET_CURR(buff), src, len);                                                                         \
        BUFF_SET_CURR(buff, BUFF_GET_CURR(buff) + len);                                                                \
    } while (0)
#define BUFF_COPY_AT(buff, offset, src, len) memcpy(BUFF_GET_HEAD(buff) + offset, src, len)
#define BUFF_SET4_MV(buff, value)                                                                                      \
    do {                                                                                                               \
        BUFF_SETV_MV(buff, U322U8A(value, 0));                                                                         \
        BUFF_SETV_MV(buff, U322U8A(value, 1));                                                                         \
        BUFF_SETV_MV(buff, U322U8A(value, 2));                                                                         \
        BUFF_SETV_MV(buff, U322U8A(value, 3));                                                                         \
    } while (0)
#define BUFF_SET4_AT(buff, offset, value)                                                                              \
    do {                                                                                                               \
        BUFF_SETV_AT(buff, offset, U322U8A(value, 0));                                                                 \
        BUFF_SETV_AT(buff, offset + 1, U322U8A(value, 1));                                                             \
        BUFF_SETV_AT(buff, offset + 2, U322U8A(value, 2));                                                             \
        BUFF_SETV_AT(buff, offset + 3, U322U8A(value, 3));                                                             \
    } while (0)
#define BUFF_GETV_MV(buff)         (*(buff).curr++)
#define BUFF_GETV_AT(buff, offset) (*((buff).head + offset))
#define BUFF_GET4_MV(buff)                                                                                             \
    ({                                                                                                                 \
        uint32_t result = U8A2U32((buff).curr);                                                                        \
        (buff).curr += 4;                                                                                              \
        result;                                                                                                        \
    })
#define BUFF_GET4_AT(buff, offset)          U8A2U32((buff).head + offset)



#define SG_INIT(sg)                       buff_t *_sg = sg
#define SG_RST(idx)                       BUFF_RST(_sg[idx])
#define SG_MOVE(idx, offset)              BUFF_MOVE(_sg[idx], offset)
#define SG_GET_HEAD(idx)                  BUFF_GET_HEAD(_sg[idx])
#define SG_GET_CURR(idx)                  BUFF_GET_CURR(_sg[idx])
#define SG_GET_SIZE(idx)                  BUFF_GET_SIZE(_sg[idx])
#define SG_SET_HEAD(idx, value)           BUFF_SET_HEAD(_sg[idx], value)
#define SG_SET_CURR(idx, value)           BUFF_SET_CURR(_sg[idx], value)
#define SG_SET_SIZE(idx, value)           BUFF_SET_SIZE(_sg[idx], value)
#define SG_OFFSET(idx)                    BUFF_OFFSET(_sg[idx])
#define SG_SETV_MV(idx, value)            BUFF_SETV_MV(_sg[idx], value)
#define SG_SETV_AT(idx, offset, value)    BUFF_SETV_AT(_sg[idx], offset, value)
#define SG_COPY_MV(idx, src, len)         BUFF_COPY_MV(_sg[idx], src, len)
#define SG_COPY_AT(idx, offset, src, len) BUFF_COPY_AT(_sg[idx], offset, src, len)
#define SG_SET4_MV(idx, value)            BUFF_SET4_MV(_sg[idx], value)
#define SG_SET4_AT(idx, offset, value)    BUFF_SET4_AT(_sg[idx], offset, value)
#define SG_GETV_MV(idx)                   BUFF_GETV_MV(_sg[idx])
#define SG_GETV_AT(idx, offset)           BUFF_GETV_AT(_sg[idx], offset)
#define SG_GET4_MV(idx)                   BUFF_GET4_MV(_sg[idx])
#define SG_GET4_AT(idx, offset)           BUFF_GET4_AT(_sg[idx], offset)



/**************************************************
 *                      LIST
 **************************************************/

typedef struct dfbc_node {
    void        *data;
    struct dfbc_node *next;
} dfbc_node_t;

typedef struct list {
	dfbc_node_t *head;
    int     iter;
    dfbc_node_t *iter_node;
} list_t;

#define LIST_INIT(list)                                                                                                \
    do {                                                                                                               \
        list = (list_t *)malloc(sizeof(list_t));                                                                       \
        list_init(list);                                                                                               \
    } while (0)

#define LIST_FREE(list, type)                                                                                          \
    do {                                                                                                               \
        while (list_len(list)) {                                                                                       \
            LIST_DELETE(list, type, 0);                                                                                \
        }                                                                                                              \
        free(list);                                                                                                    \
    } while (0)

#define LIST_APPEND(list, type, node)                                                                                  \
    do {                                                                                                               \
        type *_node = (type *)malloc(sizeof(type));                                                                    \
        memcpy(_node, node, sizeof(type));                                                                             \
        list_append(list, _node);                                                                                      \
    } while (0)

#define LIST_DELETE(list, type, index)                                                                                 \
    do {                                                                                                               \
        type *_node = NULL;                                                                                            \
        list_get(list, index, &_node);                                                                                 \
        if (_node)                                                                                                     \
            free(_node);                                                                                               \
        list_pop(list, index);                                                                                         \
    } while (0)

void list_init(list_t *);
int  list_len(list_t *);
void list_append(list_t *, void *);
int  list_pop(list_t *, int);
int  list_pop_tail(list_t *);
int  list_get(list_t *, int, void **);
int  list_set(list_t *, int, void *);
int  list_index(list_t *, void *);
int  list_iter_init(list_t *);
int  list_iter(list_t *, void **, int *);
void list_extend(list_t *, list_t *);
int  list_swap(list_t *, int, int);
void list_reverse(list_t *);
void list_cycle(list_t *, void **, int *);
int  list_from_array(list_t *, void *, int, int);

#endif
