/*!
 * \file      LmHandler.c
 *
 * \brief     Implements the LoRaMac layer handling. 
 *            Provides the possibility to register applicative packages.
 *
 * \remark    Inspired by the examples provided on the en.i-cube_lrwan fork.
 *            MCD Application Team ( STMicroelectronics International )
 *
 * \copyright Revised BSD License, see section \ref LICENSE.
 *
 * \code
 *                ______                              _
 *               / _____)             _              | |
 *              ( (____  _____ ____ _| |_ _____  ____| |__
 *               \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 *               _____) ) ____| | | || |_| ____( (___| | | |
 *              (______/|_____)_|_|_| \__)_____)\____)_| |_|
 *              (C)2013-2018 Semtech
 *
 * \endcode
 *
 * \author    Miguel Luis ( Semtech )
 */
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include "utilities.h"
#include "timer.h"
#include "Commissioning.h"
#include "NvmDataMgmt.h"
#include "radio.h"
#include "LmHandler.h"
#include "LmhPackage.h"
#include "LmhpCompliance.h"
#include "LmhpClockSync.h"
#include "LmhpRemoteMcastSetup.h"
#include "LmhpFragmentation.h"
#include "DFLora_At.h"
#include "board.h"
#include "stdio.h"
#include "comm.h"

#include "LoRaMacTest.h"

sPilotLamp_t led;

Node_t *head = NULL;
Node_t *tail = NULL;
int count = 0;

static CommissioningParams_t CommissioningParams =
{
    .IsOtaaActivation = 1,
    .DevEui = { 0 },  // Automatically filed from secure-element
    .JoinEui = { 0 }, // Automatically filed from secure-element
    .SePin = { 0 },   // Automatically filed from secure-element
    .NetworkId = LORAWAN_NETWORK_ID,
    .DevAddr = LORAWAN_DEVICE_ADDRESS,
};

static LmhPackage_t *LmHandlerPackages[PKG_MAX_NUMBER];

/*!
 * Upper layer LoRaMac parameters
 */
static LmHandlerParams_t *LmHandlerParams;

/*!
 * Upper layer callbacks
 */
static LmHandlerCallbacks_t *LmHandlerCallbacks;

/*!
 * Used to notify LmHandler of LoRaMac events
 */
static LoRaMacPrimitives_t LoRaMacPrimitives;

/*!
 * LoRaMac callbacks
 */
static LoRaMacCallback_t LoRaMacCallbacks;

static LmHandlerJoinParams_t JoinParams =
{
    .CommissioningParams = &CommissioningParams,
    .Datarate = DR_0,
    .Status = LORAMAC_HANDLER_ERROR
};

static LmHandlerTxParams_t TxParams =
{
    .CommissioningParams = &CommissioningParams,
    .MsgType = LORAMAC_HANDLER_UNCONFIRMED_MSG,
    .AckReceived = 0,
    .Datarate = DR_0,
    .UplinkCounter = 0,
    .AppData =
    {
        .Port = 0,
        .BufferSize = 0,
        .Buffer = NULL
    },
    .TxPower = TX_POWER_0,
    .Channel = 0
};

static LmHandlerRxParams_t RxParams =
{
    .CommissioningParams = &CommissioningParams,
    .Rssi = 0,
    .Snr = 0,
    .DownlinkCounter = 0,
    .RxSlot = -1
};

static LoRaMAcHandlerBeaconParams_t BeaconParams =
{
    .State = LORAMAC_HANDLER_BEACON_ACQUIRING,
    .Info =
    {
        .Time = { .Seconds = 0, .SubSeconds = 0 },
        .Frequency = 0,
        .Datarate = 0,
        .Rssi = 0,
        .Snr = 0,
        .GwSpecific =
        {
            .InfoDesc = 0,
            .Info = { 0 }
        }
    }
};

/*!
 * Indicates if a switch to Class B operation is pending or not.
 *
 * TODO: Create a new structure to store the current handler states/status
 *       and add the below variable to it.
 */
static bool IsClassBSwitchPending = false;

/*!
 * \brief   MCPS-Confirm event function
 *
 * \param   [IN] mcpsConfirm - Pointer to the confirm structure,
 *                             containing confirm attributes.
 */
static void McpsConfirm( McpsConfirm_t *mcpsConfirm );

/*!
 * \brief   MCPS-Indication event function
 *
 * \param   [IN] mcpsIndication - Pointer to the indication structure,
 *               containing indication attributes.
 */
static void McpsIndication( McpsIndication_t *mcpsIndication );

/*!
 * \brief   MLME-Confirm event function
 *
 * \param   [IN] MlmeConfirm - Pointer to the confirm structure,
 *               containing confirm attributes.
 */
static void MlmeConfirm( MlmeConfirm_t *mlmeConfirm );

/*!
 * \brief   MLME-Indication event function
 *
 * \param   [IN] mlmeIndication - Pointer to the indication structure,
 *               containing indication attributes.
 */
static void MlmeIndication( MlmeIndication_t *mlmeIndication );

/*!
 * Requests network server time update
 *
 * \retval status Returns \ref LORAMAC_HANDLER_SET if joined else \ref LORAMAC_HANDLER_RESET
 */
static LmHandlerErrorStatus_t LmHandlerDeviceTimeReq( void );

/*!
 * Starts the beacon search
 *
 * \retval status Returns \ref LORAMAC_HANDLER_SET if joined else \ref LORAMAC_HANDLER_RESET
 */
static LmHandlerErrorStatus_t LmHandlerBeaconReq( void );

/*
 *=============================================================================
 * PACKAGES HANDLING
 *=============================================================================
 */
typedef enum PackageNotifyTypes_e
{
    PACKAGE_MCPS_CONFIRM,
    PACKAGE_MCPS_INDICATION,
    PACKAGE_MLME_CONFIRM,
    PACKAGE_MLME_INDICATION,
}PackageNotifyTypes_t;

/*!
 * Notifies the package to process the LoRaMac callbacks.
 *
 * \param [IN] notifyType MAC notification type [PACKAGE_MCPS_CONFIRM,
 *                                               PACKAGE_MCPS_INDICATION,
 *                                               PACKAGE_MLME_CONFIRM,
 *                                               PACKAGE_MLME_INDICATION]
 * \param[IN] params      Notification parameters. The params type can be
 *                        [McpsConfirm_t, McpsIndication_t, MlmeConfirm_t, MlmeIndication_t]
 */
static void LmHandlerPackagesNotify( PackageNotifyTypes_t notifyType, void *params );

static void LmHandlerPackagesProcess( void );

LmHandlerErrorStatus_t LmHandlerInit( LmHandlerCallbacks_t *handlerCallbacks,
                                      LmHandlerParams_t *handlerParams )
{
//    uint16_t nbNvmData = 0;
//    LoRaMacNvmData_t *nvm = GetNvmData();
//    MibRequestConfirm_t mibReq;
    LmHandlerParams = handlerParams;
    LmHandlerCallbacks = handlerCallbacks;
    LoRaMacStatus_t status ;
    LoRaMacPrimitives.MacMcpsConfirm = McpsConfirm;
    LoRaMacPrimitives.MacMcpsIndication = McpsIndication;
    LoRaMacPrimitives.MacMlmeConfirm = MlmeConfirm;
    LoRaMacPrimitives.MacMlmeIndication = MlmeIndication;
    LoRaMacCallbacks.GetBatteryLevel = LmHandlerCallbacks->GetBatteryLevel;
    LoRaMacCallbacks.GetTemperatureLevel = LmHandlerCallbacks->GetTemperature;
    LoRaMacCallbacks.NvmDataChange  = NvmDataMgmtEvent;
    LoRaMacCallbacks.MacProcessNotify = LmHandlerCallbacks->OnMacProcess;

    IsClassBSwitchPending = false;
    //openFalse();
#include "dfbc.h"
    /*{
    char fV[20];
    floatToString(dtuConfig.loraWanPara.antennGain,fV,2);
    DBG("dtuConfig.loraWanPara.antennGain=%s\n",fV);
	}*/

   // closeFalse();
    // Restore data if required
    // nbNvmData = NvmDataMgmtRestore( );
    // printf("nbNvmData=%d\n",nbNvmData);
    // Try to restore from NVM and query the mac if possible.
    dtuConfig.loraWanPara.joinState =0;
    uint8_t sn[9]={0};
    GET_UNIQUE_ID(sn);
    md5_md5(sn,8,str);
    memcpy(sn,str,8);
    memcpy1(dtuConfig.hardPara.sn,sn,8);
    status = LoRaMacInitialization( &LoRaMacPrimitives, &LoRaMacCallbacks, LmHandlerParams->Region );

    if( status != LORAMAC_STATUS_OK )
    {
        //printf("status =%d\n",LoRaMacInitialization( &LoRaMacPrimitives, &LoRaMacCallbacks, LmHandlerParams->Region ));
       // writeLog("[ERROR] LORAMAC_STATUS =%d\r\n",status);
        return LORAMAC_HANDLER_ERROR;
    }

    MibRequestConfirm_t mibReq;
    if(dtuConfig.lmRecover) {
        dtuConfig.lmRecover = false;
        // get mib
        // mibReq.Type = MIB_DEV_EUI;
        // LoRaMacMibGetRequestConfirm( &mibReq );
        // memcpy1( CommissioningParams.DevEui, mibReq.Param.DevEui, 8 );
        // memcpy1( dtuConfig.loraWanPara.devEui, mibReq.Param.DevEui, 8 );
        mibReq.Type = MIB_JOIN_EUI;
        LoRaMacMibGetRequestConfirm( &mibReq );
        memcpy1( CommissioningParams.JoinEui, mibReq.Param.JoinEui, 8 );
        memcpy1( dtuConfig.loraWanPara.joinEui, mibReq.Param.DevEui, 8 );
        mibReq.Type = MIB_SE_PIN;
        LoRaMacMibGetRequestConfirm( &mibReq );
        memcpy1( CommissioningParams.SePin, mibReq.Param.SePin, 4 );
        // set mib
        mibReq.Type = MIB_ANTENNA_GAIN;
        mibReq.Param.AntennaGain = 3;//2.15;
        LoRaMacMibSetRequestConfirm( &mibReq );
        dtuConfig.loraWanPara.antennGain = 2.15;
        mibReq.Type = MIB_DEV_ADDR;
        mibReq.Param.DevAddr = 0xDF123456;
        LoRaMacMibSetRequestConfirm( &mibReq );
        dtuConfig.loraWanPara.devAddr = 0xDF123456;
        mibReq.Type = MIB_ADR;
        mibReq.Param.AdrEnable = false;
        LoRaMacMibSetRequestConfirm( &mibReq );
        dtuConfig.loraWanPara.adr = false;
        mibReq.Type = MIB_CHANNELS_NB_TRANS;
        mibReq.Param.ChannelsNbTrans = 1;
        LoRaMacMibSetRequestConfirm( &mibReq );
        dtuConfig.loraWanPara.nbtrials = 1;
        mibReq.Type = MIB_DEV_EUI;
        mibReq.Param.DevEui = dtuConfig.loraWanPara.devEui;
        LoRaMacMibSetRequestConfirm( &mibReq );
        memcpy1( CommissioningParams.DevEui, mibReq.Param.DevEui, 8 );


        // get key
        LoRaMacNvmData_t *nvm = GetNvmData();
        for (int i = 0; i < NUM_OF_KEYS; i++) {
            if(nvm->SecureElement.KeyList[i].KeyID == APP_KEY) {
                memcpy1(dtuConfig.loraWanPara.appKey, nvm->SecureElement.KeyList[i].KeyValue, 16);
            } else if(nvm->SecureElement.KeyList[i].KeyID == APP_S_KEY) {
                memcpy1(dtuConfig.loraWanPara.appSKey, nvm->SecureElement.KeyList[i].KeyValue, 16);
            } else if(nvm->SecureElement.KeyList[i].KeyID == NWK_S_ENC_KEY) {
                memcpy1(dtuConfig.loraWanPara.nwkSKey, nvm->SecureElement.KeyList[i].KeyValue, 16);
            }
        }
        // not supposed to take keys this way
        // memcpy1(dtuConfig.loraWanPara.appKey, nvm->SecureElement.KeyList[1].KeyValue, 16);
        // memcpy1( dtuConfig.loraWanPara.appSKey, nvm->SecureElement.KeyList[7].KeyValue, 16);
        // memcpy1( dtuConfig.loraWanPara.nwkSKey, nvm->SecureElement.KeyList[6].KeyValue, 16);
    } else {
        LmHandlerCallbacks->OnNvmDataChange( LORAMAC_HANDLER_NVM_RESTORE, sizeof(dtuConfig));
        dtuConfig.saveMask.value = 0xBFFF;
        //跳转到U盘

        //同步配置到协议栈
        syncConfig();
		//#if (ACTIVE_REGION == DF_LORAMAC_REGION_EU868)
        loadUpLinkCntAndEu8686Channel();
		//#endif
        //dtuConfig.loraWanPara.region = LmHandlerParams->Region;
        setSubBand(dtuConfig.loraWanPara.subBand);
        if(dtuConfig.logPrint == 2){
            dtuConfig.logPrint = dtuConfig.defaultLogPrint;
        }
    }

    /*
    printf("----------------------\nnbNvmData: %d\n", nbNvmData);
    if (nbNvmData > 0) {
        //从nvm读取到有用信息
        LmHandlerCallbacks->OnNvmDataChange( LORAMAC_HANDLER_NVM_RESTORE, nbNvmData );
        //loraWan所有的信息都需要重新配置
        dtuConfig.saveMask.value = 0xBFFF;
        //跳转到U盘

        //同步配置到协议栈
        openFalse();
        syncConfig();
        closeFalse();
        //dtuConfig.loraWanPara.region = LmHandlerParams->Region;
        setSubBand(dtuConfig.loraWanPara.subBand);
        if(dtuConfig.logPrint == 2){
            dtuConfig.logPrint = dtuConfig.defaultLogPrint;
        }

    } else {
        //若未从nvm读到有效数据，则初始化一部分

        sysLog("--------------crc check failure---------------\n");
       // writeLog("flashreset\r\n",0);
        dtuConfig.loopOperation.reTrans = 0;
        dtuConfig.loraWanPara.region = LmHandlerParams->Region;
        dtuConfig.cfgCnt = 0x102;
        switch(dtuConfig.loraWanPara.region){
          case LORAMAC_REGION_CN470:
              setSubBand(11);
              dtuConfig.loraWanPara.subBand = 11;
              dtuConfig.loraWanPara.eirp = 17;
            break;
          case LORAMAC_REGION_US915:
              setSubBand(2);
              dtuConfig.loraWanPara.subBand = 2;
              dtuConfig.loraWanPara.eirp = 22;
            break;
        case LORAMAC_REGION_AU915:
            setSubBand(2);
            dtuConfig.loraWanPara.subBand = 2;
            dtuConfig.loraWanPara.eirp    = 22;
            break;
        case LORAMAC_REGION_EU868:
            dtuConfig.loraWanPara.eirp = 16;
            break;
        case LORAMAC_REGION_IN865:
            dtuConfig.loraWanPara.eirp = 22;
            break;
        case LORAMAC_REGION_RU864:
            dtuConfig.loraWanPara.eirp = 16;
            break;
          default:
            break;
        }
        //产品密码
        for (int i = 0; i < 6; i++) {
               dtuConfig.hardPara.passwd[i] = '0';
        }
        for (int i = 0; i < 4; i++) {
               dtuConfig.loopOperation.io[i].enable = 0;
        }
        for (int i = 0; i < 4; i++) {
            dtuConfig.loopOperation.ioCount[i].enable = 0;
        }
        for (int i = 0; i < 4; i++) {
            dtuConfig.loopOperation.ioIn[i].enable = 0;
        }
        //产品名称
        memcpy(dtuConfig.hardPara.name,"DF-RS485",8);
        //发射功率
        //dtuConfig.loraWanPara.eirp = 5;
        //天线增益
        dtuConfig.loraWanPara.antennGain = 2.15;
        //大循环循环事件
        dtuConfig.hardPara.loopTime = 60000; //ms
        //dtuConfig.loraWanPara.dataRate = ;
        dtuConfig.loraWanPara.port = 3;

        //默认增益为2.15
        mibReq.Type = MIB_ANTENNA_GAIN;
        mibReq.Param.AntennaGain = 3;//2.15;
        LoRaMacMibSetRequestConfirm( &mibReq );

        //  printf("LoRaMacStart1\n");
        dtuConfig.hardPara.coding = 1;
        dtuConfig.hardPara.mcuSleep = 0;

        dtuConfig.loraWanPara.joinType = ACTIVATION_TYPE_OTAA;

        //是否打印log信息
        dtuConfig.logPrint = 1;


        // Read secure-element DEV_EUI, JOI_EUI and SE_PIN values.

        mibReq.Type = MIB_DEV_EUI;
        LoRaMacMibGetRequestConfirm( &mibReq );
        memcpy1( CommissioningParams.DevEui, mibReq.Param.DevEui, 8 );
        memcpy1( dtuConfig.loraWanPara.devEui, mibReq.Param.DevEui, 8 );
        mibReq.Type = MIB_JOIN_EUI;
        LoRaMacMibGetRequestConfirm( &mibReq );
        memcpy1( CommissioningParams.JoinEui, mibReq.Param.JoinEui, 8 );
        memcpy1( dtuConfig.loraWanPara.joinEui, mibReq.Param.DevEui, 8 );
        mibReq.Type = MIB_SE_PIN;
        LoRaMacMibGetRequestConfirm( &mibReq );
        memcpy1( CommissioningParams.SePin, mibReq.Param.SePin, 4 );

        mibReq.Type = MIB_DEV_ADDR;
        mibReq.Param.DevAddr = 0xDF123456;
        LoRaMacMibSetRequestConfirm( &mibReq );
        dtuConfig.loraWanPara.devAddr = 0xDF123456;
        //经纬度,海拔设置
        dtuConfig.loopOperation.coord.altitude = 500;
        dtuConfig.loopOperation.coord.latitude = 30.67;
        dtuConfig.loopOperation.coord.longitude = 104.06;

        //ADR 默认关闭
        mibReq.Type = MIB_ADR;
        mibReq.Param.AdrEnable = false;
        LoRaMacMibSetRequestConfirm( &mibReq );

        dtuConfig.loraWanPara.adr = false;
        memcpy1( dtuConfig.loraWanPara.appKey, nvm->SecureElement.KeyList[1].KeyValue, 16);
        memcpy1( dtuConfig.loraWanPara.appSKey, nvm->SecureElement.KeyList[7].KeyValue, 16);
        memcpy1( dtuConfig.loraWanPara.nwkSKey, nvm->SecureElement.KeyList[6].KeyValue, 16);
        dtuConfig.loopOperation.rs485Retry = 1;
        //应答包重传次数
        dtuConfig.loraWanPara.nbtrials = 1;
        mibReq.Type = MIB_CHANNELS_NB_TRANS;
        mibReq.Param.ChannelsNbTrans = 1;
        LoRaMacMibSetRequestConfirm( &mibReq );
        dtuConfig.hardPara.baud[0] =  115200;
        dtuConfig.hardPara.uartConfig[0].value =  3;
        
        //dtuConfig.loopOperation.rs485Baud =  115200;
        //dtuConfig.loopOperation.rs485Config.value =  3;
        
        dtuConfig.loopOperation.power[0].supplyTime = -1;
        dtuConfig.loopOperation.power[1].supplyTime = -1;
        dtuConfig.loopOperation.power[0].enable = 0;
        dtuConfig.loopOperation.power[1].enable = 0;

        printf("RECOVER --- ADN RESET\n");
        dtuConfig.loopOperation.adnDisable = false;

        dtuConfig.loraWanPara.uartPort[0] = 22;
        for(uint8_t i= 0;i<2;i++){
            dtuConfig.loopOperation.adc[i].max=5000;
            dtuConfig.loopOperation.adc[i].max=0;
        }

        dtuConfig.hardPara.addr = 0;
        deletLogFile();
    }
    */

    if(dtuConfig.loraWanPara.singleFreq == 0){

    }else{
#include "Region.h"
#include "dfbc.h"
    	ChannelAddParams_t channelAdd;
    	ChannelRemoveParams_t channelRemove;
    	ChannelParams_t newChannel;
    	DBG("\n");
    	MibRequestConfirm_t mibReq;
    	mibReq.Type = MIB_CHANNELS_MASK;
    	LoRaMacMibGetRequestConfirm( &mibReq );
    	for(uint8_t j = 0;j < 16; j++){
    	  if(((mibReq.Param.ChannelsMask[0]>>j) & 0x01)   == 0x01 ){
    			channelRemove.ChannelId = j;
    			RegionEU868ChannelsRemove(&channelRemove);
    	  }
    	}

    	newChannel.Frequency = dtuConfig.loraWanPara.singleFreq;
    	newChannel.DrRange.Fields.Min=0;
    	newChannel.DrRange.Fields.Max=5;
    	newChannel.Rx1Frequency = dtuConfig.loraWanPara.singleFreq;
    	channelAdd.ChannelId = 0;
    	channelAdd.NewChannel = &newChannel;
    	DBG("\n");
    	RegionEU868ChannelAdd( &channelAdd );
    }

    mibReq.Type = MIB_DEV_EUI;
    mibReq.Param.DevEui = dtuConfig.loraWanPara.devEui;
    LoRaMacMibSetRequestConfirm( &mibReq );
    mibReq.Type = MIB_PUBLIC_NETWORK;
    mibReq.Param.EnablePublicNetwork = LmHandlerParams->PublicNetworkEnable;
    LoRaMacMibSetRequestConfirm( &mibReq );

    LoRaMacTestSetDutyCycleOn( LmHandlerParams->DutyCycleEnabled );

    LoRaMacStart( );

    mibReq.Type = MIB_NETWORK_ACTIVATION;

    if( LoRaMacMibGetRequestConfirm( &mibReq ) == LORAMAC_STATUS_OK )
    {


        if( mibReq.Param.NetworkActivation == ACTIVATION_TYPE_NONE )
        {
            dtuConfig.loraWanPara.joinType = ACTIVATION_TYPE_OTAA;
            LmHandlerCallbacks->OnNetworkParametersChange( &CommissioningParams );
        }
    }
    
    const char *tmp;
    uint8_t str1[100]={0};
    tmp = arrtoStr(dtuConfig.loraWanPara.devEui, 8);
    sprintf(str1,"[devname] %s,%s\n",dtuConfig.hardPara.name,tmp);
    sysLog(str1);

    NvmDataMgmtEvent1(LORAMAC_NVM_NOTIFY_FLAG_MAC_GROUP2);
    return LORAMAC_HANDLER_SUCCESS;
}

bool LmHandlerIsBusy( void )
{
    if( LoRaMacIsBusy( ) == true )
    {
        sprintf(str,"busy ,state=%d\n",getMacState1());
        sysLog(str);
        //writeLog(str, getSysTime());
        if(getMacState1() == 0x02){
            if(reStartMac() == LORAMAC_STATUS_OK){
                //重置硬件
               // writeLog("\n\nre\n", getSysTime());
                SX126xIoReInit();
            }
        }else{
            reCount();
        }
        return true;
    }

    if( LmHandlerJoinStatus( ) != LORAMAC_HANDLER_SET || dtuConfig.loopOperation.fakeJoinFlag == 1 )
    {
        // The network isn't yet joined, try again later.

        LmHandlerJoin( );

        return true;
    }
    if(dtuConfig.loraWanPara.joinType == ACTIVATION_TYPE_ABP &&dtuConfig.loraWanPara.joinState ==0 ){

        MibRequestConfirm_t mibReq;
        LmHandlerJoinParams_t joinParams =
        {
            .CommissioningParams = &CommissioningParams,
            .Datarate = LmHandlerParams->TxDatarate,
            .Status = LORAMAC_HANDLER_SUCCESS
        };
        sysLog("ABP JOIN\n");
        mibReq.Type = MIB_NETWORK_ACTIVATION;
        CommissioningParams.IsOtaaActivation = false;
        mibReq.Param.NetworkActivation = ACTIVATION_TYPE_ABP;
        LoRaMacMibSetRequestConfirm( &mibReq );
        // Notify upper layer
        LmHandlerCallbacks->OnJoinRequest( &joinParams );
    }


    if( LmHandlerPackages[PACKAGE_ID_COMPLIANCE]->IsRunning( ) == true )
    {
        return true;
    }
    return false;
}

void LmHandlerProcess( void )
{
    uint16_t size = 0;

    // Process Radio IRQ
    if( Radio.IrqProcess != NULL )
    {
        Radio.IrqProcess( );
    }

    // Processes the LoRaMac events
    LoRaMacProcess( );

    // Call all packages process functions
    LmHandlerPackagesProcess( );

    // Store to NVM if required
    size = NvmDataMgmtStore( );

    if( size > 0 )
    {
        LmHandlerCallbacks->OnNvmDataChange( LORAMAC_HANDLER_NVM_STORE, size );
    }
}

void LmHandlerReJoinRequest( Mlme_t jionType){
    MlmeReq_t mlmeReq;

    mlmeReq.Type = MLME_JOIN;
    mlmeReq.Req.Join.Datarate = LmHandlerParams->TxDatarate;
    // Update commissioning parameters activation type variable.
    CommissioningParams.IsOtaaActivation = true;

    // Starts the OTAA join procedure
    LmHandlerCallbacks->OnMacMlmeRequest( LoRaMacMlmeRequest( &mlmeReq ), &mlmeReq, mlmeReq.ReqReturn.DutyCycleWaitTime );
}

/*!
 * Join a LoRa Network in classA
 *
 * \Note if the device is ABP, this is a pass through function
 *
 * \param [IN] isOtaa Indicates which activation mode must be used
 */
static void LmHandlerJoinRequest( bool isOtaa )
{
    DBG("OTAA JOIN......\n");
    if( isOtaa == true )
    {
        MlmeReq_t mlmeReq;

        mlmeReq.Type = MLME_JOIN;
        mlmeReq.Req.Join.Datarate = LmHandlerParams->TxDatarate;
        // Update commissioning parameters activation type variable.
        CommissioningParams.IsOtaaActivation = true;

        // Starts the OTAA join procedure
        LmHandlerCallbacks->OnMacMlmeRequest( LoRaMacMlmeRequest( &mlmeReq ), &mlmeReq, mlmeReq.ReqReturn.DutyCycleWaitTime );
        data_blink(1000);
    } else {
        MibRequestConfirm_t mibReq;
        LmHandlerJoinParams_t joinParams =
        {
            .CommissioningParams = &CommissioningParams,
            .Datarate = LmHandlerParams->TxDatarate,
            .Status = LORAMAC_HANDLER_SUCCESS
        };
        sysLog("ABP JOIN\n");
        mibReq.Type = MIB_NETWORK_ACTIVATION;
        mibReq.Param.NetworkActivation = ACTIVATION_TYPE_ABP;
        LoRaMacMibSetRequestConfirm( &mibReq );

        // Notify upper layer
        LmHandlerCallbacks->OnJoinRequest( &joinParams );
    }
}

void LmHandlerJoin( void )
{
    LmHandlerJoinRequest( CommissioningParams.IsOtaaActivation );
}

LmHandlerFlagStatus_t LmHandlerJoinStatus( void )
{
    MibRequestConfirm_t mibReq;
    LoRaMacStatus_t status;

    mibReq.Type = MIB_NETWORK_ACTIVATION;
    status = LoRaMacMibGetRequestConfirm( &mibReq );

    if( status == LORAMAC_STATUS_OK )
    {
        if( mibReq.Param.NetworkActivation == ACTIVATION_TYPE_NONE )
        {
            return LORAMAC_HANDLER_RESET;
        }
        else
        {
            return LORAMAC_HANDLER_SET;
        }
    }
    else
    {
        return LORAMAC_HANDLER_RESET;
    }
}

LmHandlerErrorStatus_t LmHandlerSend( LmHandlerAppData_t *appData, LmHandlerMsgTypes_t isTxConfirmed )
{
    LoRaMacStatus_t status;
    McpsReq_t mcpsReq;
    LoRaMacTxInfo_t txInfo;
    appData->Port = 2;
    if( LmHandlerJoinStatus( ) != LORAMAC_HANDLER_SET )
    {
        // The network isn't joined, try again.
    	//printf("%s():%d\n",__FUNCTION__,__LINE__);
        LmHandlerJoinRequest( CommissioningParams.IsOtaaActivation );
        return LORAMAC_HANDLER_ERROR;
    }

    if( ( LmHandlerPackages[PACKAGE_ID_COMPLIANCE]->IsRunning( ) == true ) && ( appData->Port != LmHandlerPackages[PACKAGE_ID_COMPLIANCE]->Port ) && ( appData->Port != 0 ) )
    {
    	//printf("%s():%d\n",__FUNCTION__,__LINE__);
        return LORAMAC_HANDLER_ERROR;
    }
    LoRaMacNvmData_t *nvm = GetNvmData();
    //isTxConfirmed
    //TxParams.MsgType = dtuConfig.loraWanPara.isTxConfirmed;
    TxParams.MsgType = isTxConfirmed;
   // isTxConfirmed = ( TxParams.MsgType == LORAMAC_HANDLER_UNCONFIRMED_MSG ) ? MCPS_UNCONFIRMED : MCPS_CONFIRMED;

    //printf("%s():%d\n",__FUNCTION__,__LINE__);
    mcpsReq.Req.Unconfirmed.Datarate = nvm->MacGroup1.ChannelsDatarate;
    if( LoRaMacQueryTxPossible( appData->BufferSize, &txInfo ) != LORAMAC_STATUS_OK ){
        // Send empty frame in order to flush MAC commands
    	//printf("%s():%d\n",__FUNCTION__,__LINE__);
        TxParams.MsgType = LORAMAC_HANDLER_UNCONFIRMED_MSG;
        mcpsReq.Type = MCPS_UNCONFIRMED;
        mcpsReq.Req.Unconfirmed.fBuffer = NULL;
        mcpsReq.Req.Unconfirmed.fBufferSize = 0;
    } else {
    	//printf("%s():%d\n",__FUNCTION__,__LINE__);
        if(dtuConfig.loraWanPara.uartMode == 0x00){
          appData->Port =  dtuConfig.loraWanPara.port;
          mcpsReq.Req.Unconfirmed.fPort =  dtuConfig.loraWanPara.port;
        }else if (dtuConfig.loraWanPara.uartMode == 0x01){
        	//printf("%s():%d\n",__FUNCTION__,__LINE__);
            appData->Port =  dtuConfig.loraWanPara.uartPort[0];
            mcpsReq.Req.Unconfirmed.fPort =  dtuConfig.loraWanPara.uartPort[0];
        }else{
        	//printf("%s():%d\n",__FUNCTION__,__LINE__);
        }
        //printf("%s():%d\n",__FUNCTION__,__LINE__);
        mcpsReq.Req.Unconfirmed.fBufferSize = appData->BufferSize;
        mcpsReq.Req.Unconfirmed.fBuffer = appData->Buffer;
        if( TxParams.MsgType == LORAMAC_HANDLER_UNCONFIRMED_MSG ) {
            mcpsReq.Type = MCPS_UNCONFIRMED;
            //printf("%s():%d\n",__FUNCTION__,__LINE__);
        } else {
        	//printf("%s():%d\n",__FUNCTION__,__LINE__);
            mcpsReq.Type = MCPS_CONFIRMED;
            mcpsReq.Req.Confirmed.NbTrials = 3;
        }
    }
    //printf("%s():%d\n",__FUNCTION__,__LINE__);
    TxParams.AppData = *appData;
    //DBG("---------TxParams.AppData = %x %x %x\n",TxParams.AppData.Buffer[0],TxParams.AppData.Buffer[1],TxParams.AppData.Buffer[2]);
    TxParams.Datarate = nvm->MacGroup1.ChannelsDatarate;
    //printf("%s():%d\n",__FUNCTION__,__LINE__);
    status = LoRaMacMcpsRequest( &mcpsReq );

    if(status != LORAMAC_STATUS_OK){
    	//printf("%s():%d\n",__FUNCTION__,__LINE__);
        sprintf(str,"LoRa send status=%d\r\n",status);
        if(status == 11){
        	//printf("%s():%d\n",__FUNCTION__,__LINE__);
          // sprintf(str+strlen(str)-2,"  dr=%d\n",nvm->MacGroup1.ChannelsDatarate);
        }
        //printf("%s():%d\n",__FUNCTION__,__LINE__);
        sysLog(str);
       // writeLog(str,getSysTime());
    }

    //printf("%s():%d mcpsReq.ReqReturn.DutyCycleWaitTime=%d\n",__FUNCTION__,__LINE__,mcpsReq.ReqReturn.DutyCycleWaitTime);
    LmHandlerCallbacks->OnMacMcpsRequest( status, &mcpsReq, mcpsReq.ReqReturn.DutyCycleWaitTime );
    //printf("%s():%d\n",__FUNCTION__,__LINE__);
    if( status == LORAMAC_STATUS_OK )
    {
    	//printf("%s():%d\n",__FUNCTION__,__LINE__);
        return LORAMAC_HANDLER_SUCCESS;
    } else {
        if(status != 11){
        	//printf("%s():%d\n",__FUNCTION__,__LINE__);
           return LORAMAC_HANDLER_ERROR;
        }else{
        	//printf("%s():%d\n",__FUNCTION__,__LINE__);
            return LORAMAC_HANDLER_DUTY;
        }
    }
}

static LmHandlerErrorStatus_t LmHandlerDeviceTimeReq( void )
{
    LoRaMacStatus_t status;
    MlmeReq_t mlmeReq;

    mlmeReq.Type = MLME_DEVICE_TIME;

    status = LoRaMacMlmeRequest( &mlmeReq );
    LmHandlerCallbacks->OnMacMlmeRequest( status, &mlmeReq, mlmeReq.ReqReturn.DutyCycleWaitTime );

    if( status == LORAMAC_STATUS_OK )
    {
        return LORAMAC_HANDLER_SUCCESS;
    }
    else
    {
        return LORAMAC_HANDLER_ERROR;
    }
}

static LmHandlerErrorStatus_t LmHandlerBeaconReq( void )
{
    LoRaMacStatus_t status;
    MlmeReq_t mlmeReq;

    mlmeReq.Type = MLME_BEACON_ACQUISITION;

    status = LoRaMacMlmeRequest( &mlmeReq );
    LmHandlerCallbacks->OnMacMlmeRequest( status, &mlmeReq, mlmeReq.ReqReturn.DutyCycleWaitTime );

    if( status == LORAMAC_STATUS_OK )
    {
        return LORAMAC_HANDLER_SUCCESS;
    }
    else
    {
        return LORAMAC_HANDLER_ERROR;
    }
}

LmHandlerErrorStatus_t LmHandlerPingSlotReq( uint8_t periodicity )
{
    LoRaMacStatus_t status;
    MlmeReq_t mlmeReq;

    mlmeReq.Type = MLME_PING_SLOT_INFO;
    mlmeReq.Req.PingSlotInfo.PingSlot.Fields.Periodicity = periodicity;
    mlmeReq.Req.PingSlotInfo.PingSlot.Fields.RFU = 0;

    status = LoRaMacMlmeRequest( &mlmeReq );
    LmHandlerCallbacks->OnMacMlmeRequest( status, &mlmeReq, mlmeReq.ReqReturn.DutyCycleWaitTime );

    if( status == LORAMAC_STATUS_OK )
    {
        // Send an empty message
        LmHandlerAppData_t appData =
        {
            .Buffer = NULL,
            .BufferSize = 0,
            .Port = 0
        };
        return LmHandlerSend( &appData, LORAMAC_HANDLER_UNCONFIRMED_MSG );
    }
    else
    {
        return LORAMAC_HANDLER_ERROR;
    }
}

LmHandlerErrorStatus_t LmHandlerRequestClass( DeviceClass_t newClass )
{
    MibRequestConfirm_t mibReq;
    DeviceClass_t currentClass;
    LmHandlerErrorStatus_t errorStatus = LORAMAC_HANDLER_SUCCESS;

    mibReq.Type = MIB_DEVICE_CLASS;
    LoRaMacMibGetRequestConfirm( &mibReq );
    currentClass = mibReq.Param.Class;
    //printf("currentClass=%d\n",currentClass);
    // Attempt to switch only if class update
    if( currentClass != newClass )
    {
        switch( newClass )
        {
        case CLASS_A:
            {
                if( currentClass != CLASS_A )
                {
                    mibReq.Param.Class = CLASS_A;
                    if( LoRaMacMibSetRequestConfirm( &mibReq ) == LORAMAC_STATUS_OK )
                    {
                        // Switch is instantaneous
                        LmHandlerCallbacks->OnClassChange( CLASS_A );
                    }
                    else
                    {
                        errorStatus = LORAMAC_HANDLER_ERROR;
                    }
                }
            }
            break;
        case CLASS_B:
            {
                if( currentClass != CLASS_A )
                {
                    errorStatus = LORAMAC_HANDLER_ERROR;
                }
                // Beacon must first be acquired
                errorStatus = LmHandlerDeviceTimeReq( );
                IsClassBSwitchPending = true;
            }
            break;
        case CLASS_C:
            {
                if( currentClass != CLASS_A )
                {
                    errorStatus = LORAMAC_HANDLER_ERROR;
                }
                // Switch is instantaneous
                mibReq.Param.Class = CLASS_C;
                if( LoRaMacMibSetRequestConfirm( &mibReq ) == LORAMAC_STATUS_OK )
                {
                    LmHandlerCallbacks->OnClassChange( CLASS_C );
                }
                else
                {
                    errorStatus = LORAMAC_HANDLER_ERROR;
                }
            }
            break;
        default:
            break;
        }
    }
    return errorStatus;
}

DeviceClass_t LmHandlerGetCurrentClass( void )
{
    MibRequestConfirm_t mibReq;

    mibReq.Type = MIB_DEVICE_CLASS;
    LoRaMacMibGetRequestConfirm( &mibReq );

    return mibReq.Param.Class;
}

int8_t LmHandlerGetCurrentDatarate( void )
{
    MibRequestConfirm_t mibGet;

    mibGet.Type = MIB_CHANNELS_DATARATE;
    LoRaMacMibGetRequestConfirm( &mibGet );

    return mibGet.Param.ChannelsDatarate;
}

LoRaMacRegion_t LmHandlerGetActiveRegion( void )
{
    return LmHandlerParams->Region;
}

LmHandlerErrorStatus_t LmHandlerSetSystemMaxRxError( uint32_t maxErrorInMs )
{
    MibRequestConfirm_t mibReq;

    mibReq.Type = MIB_SYSTEM_MAX_RX_ERROR;
    mibReq.Param.SystemMaxRxError = maxErrorInMs;
    if( LoRaMacMibSetRequestConfirm( &mibReq ) != LORAMAC_STATUS_OK )
    {
        return LORAMAC_HANDLER_ERROR;
    }
    return LORAMAC_HANDLER_SUCCESS;
}

/*
 *=============================================================================
 * LORAMAC NOTIFICATIONS HANDLING
 *=============================================================================
 */

static void McpsConfirm( McpsConfirm_t *mcpsConfirm )
{
    TxParams.IsMcpsConfirm = 1;
    TxParams.Status = mcpsConfirm->Status;
    TxParams.Datarate = mcpsConfirm->Datarate;
    TxParams.UplinkCounter = mcpsConfirm->UpLinkCounter;
    TxParams.TxPower = mcpsConfirm->TxPower;
    TxParams.Channel = mcpsConfirm->Channel;
    TxParams.AckReceived = mcpsConfirm->AckReceived;

    LmHandlerCallbacks->OnTxData( &TxParams );

    LmHandlerPackagesNotify( PACKAGE_MCPS_CONFIRM, mcpsConfirm );
}

static void McpsIndication( McpsIndication_t *mcpsIndication )
{
    LmHandlerAppData_t appData;

    RxParams.IsMcpsIndication = 1;
    RxParams.Status = mcpsIndication->Status;

    if( RxParams.Status != LORAMAC_EVENT_INFO_STATUS_OK )
    {
        return;
    }

    RxParams.Datarate = mcpsIndication->RxDatarate;
    RxParams.Rssi = mcpsIndication->Rssi;
    RxParams.Snr = mcpsIndication->Snr;
    RxParams.DownlinkCounter = mcpsIndication->DownLinkCounter;
    RxParams.RxSlot = mcpsIndication->RxSlot;

    appData.Port = mcpsIndication->Port;
    appData.BufferSize = mcpsIndication->BufferSize;
    appData.Buffer = mcpsIndication->Buffer;

    LmHandlerCallbacks->OnRxData( &appData, &RxParams );

    if( mcpsIndication->DeviceTimeAnsReceived == true )
    {
#if( LMH_SYS_TIME_UPDATE_NEW_API == 1 )
        // Provide fix values. DeviceTimeAns is accurate
        LmHandlerCallbacks->OnSysTimeUpdate( true, 0 );
#else
        LmHandlerCallbacks->OnSysTimeUpdate( );
#endif
    }
    // Call packages RxProcess function

    if (mcpsIndication->McpsIndication == MCPS_PROPRIETARY)
    {
        OnReceiveProprietary(mcpsIndication);
    }

    LmHandlerPackagesNotify( PACKAGE_MCPS_INDICATION, mcpsIndication );

    if( ( mcpsIndication->FramePending == true ) && ( LmHandlerGetCurrentClass( ) == CLASS_A ) )
    {
        // The server signals that it has pending data to be sent.
        // We schedule an uplink as soon as possible to flush the server.

        // Send an empty message
        LmHandlerAppData_t appData =
        {
            .Buffer = NULL,
            .BufferSize = 0,
            .Port = 0
        };
        //printf("McpsIndication\r\n");
        LmHandlerSend( &appData, LORAMAC_HANDLER_UNCONFIRMED_MSG );
    }
}

static void MlmeConfirm( MlmeConfirm_t *mlmeConfirm )
{
    TxParams.IsMcpsConfirm = 0;
    TxParams.Status = mlmeConfirm->Status;
    LmHandlerCallbacks->OnTxData( &TxParams );

    LmHandlerPackagesNotify( PACKAGE_MLME_CONFIRM, mlmeConfirm );

    switch( mlmeConfirm->MlmeRequest )
    {
    case MLME_JOIN:
        {
            MibRequestConfirm_t mibReq;
            mibReq.Type = MIB_DEV_ADDR;
            LoRaMacMibGetRequestConfirm( &mibReq );
            JoinParams.CommissioningParams->DevAddr = mibReq.Param.DevAddr;
            JoinParams.Datarate = LmHandlerGetCurrentDatarate( );

            if( mlmeConfirm->Status == LORAMAC_EVENT_INFO_STATUS_OK )
            {
                // Status is OK, node has joined the network
                JoinParams.Status = LORAMAC_HANDLER_SUCCESS;
            }
            else
            {
                // Join was not successful. Try to join again
                JoinParams.Status = LORAMAC_HANDLER_ERROR;
            }
            // Notify upper layer
            LmHandlerCallbacks->OnJoinRequest( &JoinParams );
        }
        break;
    case MLME_LINK_CHECK:
        {
            // Check DemodMargin
            // Check NbGateways
        }
        break;
    case MLME_DEVICE_TIME:
        {
            if( IsClassBSwitchPending == true )
            {
                LmHandlerBeaconReq( );
            }
        }
        break;
    case MLME_BEACON_ACQUISITION:
        {
            if( mlmeConfirm->Status == LORAMAC_EVENT_INFO_STATUS_OK )
            {
                // Beacon has been acquired
                // Request server for ping slot
                LmHandlerPingSlotReq( 0 );
            }
            else
            {
                // Beacon not acquired
                // Request Device Time again.
                LmHandlerDeviceTimeReq( );
            }
        }
        break;
    case MLME_PING_SLOT_INFO:
        {
            if( mlmeConfirm->Status == LORAMAC_EVENT_INFO_STATUS_OK )
            {
                MibRequestConfirm_t mibReq;

                // Class B is now activated
                mibReq.Type = MIB_DEVICE_CLASS;
                mibReq.Param.Class = CLASS_B;
                LoRaMacMibSetRequestConfirm( &mibReq );
                // Notify upper layer
                LmHandlerCallbacks->OnClassChange( CLASS_B );
                IsClassBSwitchPending = false;
            }
            else
            {
                LmHandlerPingSlotReq( 0 );
            }
        }
        break;
    default:
        break;
    }

}

static void MlmeIndication( MlmeIndication_t *mlmeIndication )
{
    RxParams.IsMcpsIndication = 0;
    RxParams.Status = mlmeIndication->Status;
    if( RxParams.Status != LORAMAC_EVENT_INFO_STATUS_BEACON_LOCKED )
    {
        LmHandlerCallbacks->OnRxData( NULL, &RxParams );
    }

    // Call packages RxProcess function
    LmHandlerPackagesNotify( PACKAGE_MLME_INDICATION, mlmeIndication );

    switch( mlmeIndication->MlmeIndication )
    {
    case MLME_SCHEDULE_UPLINK:
        {// The MAC signals that we shall provide an uplink as soon as possible
            // Send an empty message
            LmHandlerAppData_t appData =
            {
                .Buffer = NULL,
                .BufferSize = 0,
                .Port = 0
            };

            if( LmHandlerPackages[PACKAGE_ID_COMPLIANCE]->IsRunning( ) == false )
            {

                if( LoRaMacIsBusy( ) == false )
                {

                    //sysLog("busy\n");
                    //return true;

                    LmHandlerSend( &appData, LORAMAC_HANDLER_UNCONFIRMED_MSG );
                    //printf("MLME_SCHEDULE_UPLINK1\n");
                }


            }
        }
        break;
    case MLME_BEACON_LOST:
        {
            MibRequestConfirm_t mibReq;
            // Switch to class A again
            mibReq.Type = MIB_DEVICE_CLASS;
            mibReq.Param.Class = CLASS_A;
            LoRaMacMibSetRequestConfirm( &mibReq );

            BeaconParams.State = LORAMAC_HANDLER_BEACON_LOST;
            BeaconParams.Info.Time.Seconds = 0;
            BeaconParams.Info.GwSpecific.InfoDesc = 0;
            memset1( BeaconParams.Info.GwSpecific.Info, 0, 6 );

            LmHandlerCallbacks->OnClassChange( CLASS_A );
            LmHandlerCallbacks->OnBeaconStatusChange( &BeaconParams );

            LmHandlerDeviceTimeReq( );
        }
        break;
    case MLME_BEACON:
    {
        if( mlmeIndication->Status == LORAMAC_EVENT_INFO_STATUS_BEACON_LOCKED )
        {
            BeaconParams.State = LORAMAC_HANDLER_BEACON_RX;
            BeaconParams.Info = mlmeIndication->BeaconInfo;

            LmHandlerCallbacks->OnBeaconStatusChange( &BeaconParams );
        }
        else
        {
            BeaconParams.State = LORAMAC_HANDLER_BEACON_NRX;
            BeaconParams.Info = mlmeIndication->BeaconInfo;

            LmHandlerCallbacks->OnBeaconStatusChange( &BeaconParams );
        }
        break;
    }
    default:
        break;
    }
}

/*
 *=============================================================================
 * PACKAGES HANDLING
 *=============================================================================
 */

LmHandlerErrorStatus_t LmHandlerPackageRegister( uint8_t id, void *params )
{
    LmhPackage_t *package = NULL;
    switch( id )
    {
        case PACKAGE_ID_COMPLIANCE:
        {
            package = LmphCompliancePackageFactory( );
            break;
        }
        case PACKAGE_ID_CLOCK_SYNC:
        {
            package = LmphClockSyncPackageFactory( );
            break;
        }
        case PACKAGE_ID_REMOTE_MCAST_SETUP:
        {
            package = LmhpRemoteMcastSetupPackageFactory( );
            break;
        }
        case PACKAGE_ID_FRAGMENTATION:
        {
            package = LmhpFragmentationPackageFactory( );
            break;
        }
    }
    if( package != NULL )
    {
        LmHandlerPackages[id] = package;
        LmHandlerPackages[id]->OnMacMcpsRequest = LmHandlerCallbacks->OnMacMcpsRequest;
        LmHandlerPackages[id]->OnMacMlmeRequest = LmHandlerCallbacks->OnMacMlmeRequest;
        LmHandlerPackages[id]->OnJoinRequest = LmHandlerJoinRequest;
        LmHandlerPackages[id]->OnSendRequest = LmHandlerSend;
        LmHandlerPackages[id]->OnDeviceTimeRequest = LmHandlerDeviceTimeReq;
        LmHandlerPackages[id]->OnSysTimeUpdate = LmHandlerCallbacks->OnSysTimeUpdate;
        LmHandlerPackages[id]->Init( params, LmHandlerParams->DataBuffer, LmHandlerParams->DataBufferMaxSize );

        return LORAMAC_HANDLER_SUCCESS;

        return LORAMAC_HANDLER_SUCCESS;
    }
    else
    {
        return LORAMAC_HANDLER_ERROR;
    }
}

bool LmHandlerPackageIsInitialized( uint8_t id )
{
    if( LmHandlerPackages[id]->IsInitialized != NULL )
    {
        return LmHandlerPackages[id]->IsInitialized( );
    }
    else
    {
        return false;
    }
}

bool LmHandlerPackageIsRunning( uint8_t id )
{
    if( LmHandlerPackages[id]->IsRunning != NULL )
    {
        return LmHandlerPackages[id]->IsRunning( );
    }
    else
    {
        return false;
    }
}

static void LmHandlerPackagesNotify( PackageNotifyTypes_t notifyType, void *params )
{
    for( int8_t i = 0; i < PKG_MAX_NUMBER; i++ )
    {
        if( LmHandlerPackages[i] != NULL )
        {
            switch( notifyType )
            {
                case PACKAGE_MCPS_CONFIRM:
                {
                    if( LmHandlerPackages[i]->OnMcpsConfirmProcess != NULL )
                    {
                        LmHandlerPackages[i]->OnMcpsConfirmProcess( ( McpsConfirm_t* ) params );
                    }
                    break;
                }
                case PACKAGE_MCPS_INDICATION:
                {
                    if( ( LmHandlerPackages[i]->OnMcpsIndicationProcess != NULL ) &&
                        ( LmHandlerPackages[i]->Port == ( ( McpsIndication_t* )params )->Port ) )
                    {
                        LmHandlerPackages[i]->OnMcpsIndicationProcess( ( McpsIndication_t* )params );
                    }
                    break;
                }
                case PACKAGE_MLME_CONFIRM:
                {
                    if( LmHandlerPackages[i]->OnMlmeConfirmProcess != NULL )
                    {
                        LmHandlerPackages[i]->OnMlmeConfirmProcess( ( MlmeConfirm_t* )params );
                    }
                    break;
                }
                case PACKAGE_MLME_INDICATION:
                {
                    if( LmHandlerPackages[i]->OnMlmeIndicationProcess != NULL )
                    {
                        LmHandlerPackages[i]->OnMlmeIndicationProcess( params );
                    }
                    break;
                }
            }
        }
    }
}

static void LmHandlerPackagesProcess( void )
{
    for( int8_t i = 0; i < PKG_MAX_NUMBER; i++ )
    {
        if( ( LmHandlerPackages[i] != NULL ) &&
            ( LmHandlerPackages[i]->Process != NULL ) &&
            ( LmHandlerPackageIsInitialized( i ) != false ) )
        {
            LmHandlerPackages[i]->Process( );
        }
    }
}
void addNode(uint8_t *data, uint8_t dataLen, uint8_t rssi, uint8_t snr, uint8_t port) {
    Node_t *newNode = (Node_t *)malloc(sizeof(Node_t));
    newNode->data = (uint8_t *)malloc(dataLen * sizeof(uint8_t));
    for (int i = 0; i < dataLen; i++) {
        newNode->data[i] = data[i];
    }
    newNode->dataLen = dataLen;
    newNode->rssi = rssi;
    newNode->snr = snr;
    newNode->port = port;
    newNode->next = NULL;
    if (tail == NULL) {
        head = newNode;
        tail = newNode;
    } else {
        tail->next = newNode;
        tail = newNode;
    }
    if (tail - head >= 10) {
        Node_t *temp = head;
        head = head->next;
        free(temp->data);
        free(temp);
    }
}
Node_t * readNode() {
    if (head == NULL) {
        printf("The list is empty!\n");
        return NULL;
    } else {
        printf("The data in the first node is:\n");
        printf("Data: ");
        for (int i = 0; i < head->dataLen; i++) {
            printf("%d ", head->data[i]);
        }
        printf("\n");
        printf("RSSI: %d\n", head->rssi);
        printf("SNR: %d\n", head->snr);
        printf("Port: %d\n", head->port);
        Node_t *temp = head;
        head = head->next;
        //free(temp->data);
        //free(temp);
        return temp;
    }
}
