/*!
 * \file      NvmDataMgmt.c
 *
 * \brief     NVM context management implementation
 *
 * \copyright Revised BSD License, see section \ref LICENSE.
 *
 * \code
 *                ______                              _
 *               / _____)             _              | |
 *              ( (____  _____ ____ _| |_ _____  ____| |__
 *               \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 *               _____) ) ____| | | || |_| ____( (___| | | |
 *              (______/|_____)_|_|_| \__)_____)\____)_| |_|
 *              (C)2013-2017 Semtech
 *
 *               ___ _____ _   ___ _  _____ ___  ___  ___ ___
 *              / __|_   _/_\ / __| |/ / __/ _ \| _ \/ __| __|
 *              \__ \ | |/ _ \ (__| ' <| _| (_) |   / (__| _|
 *              |___/ |_/_/ \_\___|_|\_\_| \___/|_|_\\___|___|
 *              embedded.connectivity.solutions===============
 *
 * \endcode
 *
 * \author    Miguel Luis ( Semtech )
 *
 * \author    Gregory Cristian ( Semtech )
 *
 * \author    Daniel Jaeckle ( STACKFORCE )
 *
 * \author    Johannes Bruder ( STACKFORCE )
 */

#include <stdio.h>
#include "utilities.h"
#include "nvmm.h"
#include "LoRaMac.h"
#include "NvmDataMgmt.h"
#include "DFLora_At.h"
#include "dfbc.h"
#include "comm.h"

/*!
 * Enables/Disables the context storage management storage.
 * Must be enabled for LoRaWAN 1.0.4 or later.
 */
#ifndef CONTEXT_MANAGEMENT_ENABLED
#define CONTEXT_MANAGEMENT_ENABLED         1
#endif

static uint16_t NvmNotifyFlags = 0;
static uint16_t NvmNotifyFlag1 = 0;
void NvmDataMgmtEvent(uint16_t notifyFlags) {
    // printf("123456\n");
    NvmNotifyFlags = notifyFlags;
}
void NvmDataMgmtEvent1(uint16_t notifyFlags) {
    // printf("123456\n");
    setMacDone();
    NvmNotifyFlag1 = notifyFlags;
}

uint16_t calculateCRC(const uint8_t *data, size_t length) {
    uint16_t crc = 0xFFFF;
    size_t i, j;
    for (i = 0; i < length; ++i) {
        crc ^= data[i];
        for (j = 0; j < 8; ++j) {
            if (crc & 1) {
                crc >>= 1;
                crc ^= 0xA001; // 使用CRC-16/Modbus多项式
            } else {
                crc >>= 1;
            }
        }
    }
    return crc;
}

//存储用户配置数据 dtuConfig
uint16_t NvmStoreDTUConfig( void ){
    uint16_t dataSize = 0;
    MibRequestConfirm_t mibReq;
    if (NvmNotifyFlag1 == LORAMAC_NVM_NOTIFY_FLAG_NONE) {
        return 0;
    }
    uint16_t crcTemp  = dtuConfig.crc;
    dtuConfig.crc     = 0;
    uint16_t calc_crc = calculateCRC((uint8_t *)&dtuConfig, sizeof(sDtuConfig_t));
    // 与存储的crc比较
    if (calc_crc != dtuConfig.crc) {
        if (NvmNotifyFlag1 == LORAMAC_NVM_NOTIFY_FLAG_MAC_GROUP2) {
            // 若数据有改动,则从新赋值crc,再计算出一个新得分crc值
            printf("----------store dtu config-----calc_crc=%x--------\n", calc_crc);
            // dtuConfig.crc = crc;
            // crc = calculateCRC((uint8_t*)&dtuConfig,sizeof(sDtuConfig_t));
            // 把新的crc值再填入配置，写进flash
            dtuConfig.crc = calc_crc;

            NvmmWrite((uint8_t *)&dtuConfig, sizeof(sDtuConfig_t), DF_BOOT_SIZE);
            // NvmmWrite(&dtuConfig.crc,2,sizeof(sDtuConfig_t)-2);

            dataSize = 1;
        }
    } else {
        dtuConfig.crc = crcTemp;
    }
    NvmNotifyFlag1 = LORAMAC_NVM_NOTIFY_FLAG_NONE;
    return dataSize;
}

uint16_t NvmDataMgmtStore(void) {
#if (CONTEXT_MANAGEMENT_ENABLED == 1)
    uint16_t offset   = 0;
    uint16_t dataSize = 0;
    MibRequestConfirm_t mibReq;
    mibReq.Type = MIB_NVM_CTXS;
    LoRaMacMibGetRequestConfirm(&mibReq);
    LoRaMacNvmData_t *nvm = mibReq.Param.Contexts;

    // Input checks
    if (NvmNotifyFlags == LORAMAC_NVM_NOTIFY_FLAG_NONE) {
        // There was no update.

        return 0;
    }
    if (NvmNotifyFlag1 == LORAMAC_NVM_NOTIFY_FLAG_NONE) {
        // There was no update.

        return 0;
    }
    if (LoRaMacStop() != LORAMAC_STATUS_OK) {
        // printf("busy----\n");
        return 0;
    }

    mibReq.Type = MIB_DEV_ADDR;
    LoRaMacMibGetRequestConfirm( &mibReq );
    dtuConfig.loraWanPara.devAddr = mibReq.Param.DevAddr;

   // mibReq.Type = MIB_APP_S_KEY;
   // LoRaMacMibGetRequestConfirm( &mibReq );
   // mibReq.Param.DevAddr = dtuConfig.loraWanPara.appSKey;

   // mibReq.Type = MIB_NWK_KEY;
   // LoRaMacMibGetRequestConfirm( &mibReq );
   // mibReq.Param.DevAddr = dtuConfig.loraWanPara.nwkSKey;

    mibReq.Type = MIB_CHANNELS_MASK;
    LoRaMacMibGetRequestConfirm( &mibReq );
    DBG("\n");
    mibReq.Param.ChannelsMask = dtuConfig.loraWanPara.channelsMask;

    mibReq.Type = MIB_NET_ID;
    LoRaMacMibGetRequestConfirm( &mibReq );
    dtuConfig.loraWanPara.netId = mibReq.Param.NetID;

   // mibReq.Type = MIB_ADR;
    //LoRaMacMibGetRequestConfirm( &mibReq );
    //dtuConfig.loraWanPara.netId = mibReq.Param.NetID;
    //首先计算出一个现在配置的crc

    uint16_t crcTemp = dtuConfig.crc;

    dtuConfig.crc = 0;

    uint16_t crc = calculateCRC((uint8_t*)&dtuConfig,sizeof(sDtuConfig_t));

    //与存储的crc比较
    if(crc != crcTemp && NvmNotifyFlag1 == LORAMAC_NVM_NOTIFY_FLAG_MAC_GROUP2){
      //若数据有改动,则从新赋值crc,再计算出一个新得分crc值

      //dtuConfig.crc = crc;
      //crc = calculateCRC((uint8_t*)&dtuConfig,sizeof(sDtuConfig_t));
      //把新的crc值再填入配置，写进flash
      dtuConfig.crc = crc;

      NvmmWrite(&dtuConfig,sizeof(sDtuConfig_t),DF_BOOT_SIZE);
    //NvmmWrite(&dtuConfig.crc,2,sizeof(sDtuConfig_t)-2);

      dataSize = 1;
      //sprintf(str,"-----------store-------------%x\n",crc);

    //  sysLog("-----------done-------------\n");
    // Resume LoRaMac

    }else{
      dtuConfig.crc = crcTemp;
    }
    LoRaMacStart();
    NvmNotifyFlag1 = LORAMAC_NVM_NOTIFY_FLAG_NONE;
    NvmNotifyFlags = LORAMAC_NVM_NOTIFY_FLAG_NONE;
    return dataSize;
#else
    return 0;
#endif
}

static void recoverDtuConfig(void) {
    DBG("\n*************************\n*************************\nrecovery to default "
        "config\n*************************\n*************************\n");
    // crc fail, recover config from default configuration
#if (ACTIVE_REGION == DF_LORAMAC_REGION_EU868)
      DBG("EU868\n");
      dtuConfig.loraWanPara.region = LORAMAC_REGION_EU868;
      strcpy(dtuConfig.hardPara.name, "DTU-868");
      strcpy(dtuConfig.hardPara.product, "DTU-868");
#elif (ACTIVE_REGION == DF_LORAMAC_REGION_US915)
      DBG("US915\n");
      dtuConfig.loraWanPara.region = LORAMAC_REGION_US915;
      strcpy(dtuConfig.hardPara.name, "DTU-915");
      strcpy(dtuConfig.hardPara.product, "DTU-915");
#endif
    dtuConfig.cfgCnt             = 0x102;
    dtuConfig.hardPara.loss      = 4;
    dtuConfig.lmRecover       = true;
    switch (dtuConfig.loraWanPara.region) {
    case LORAMAC_REGION_CN470:
        setSubBand(11);
        dtuConfig.loraWanPara.subBand = 11;
        dtuConfig.loraWanPara.eirp    = 17;
        break;
    case LORAMAC_REGION_US915:
        setSubBand(2);
        dtuConfig.loraWanPara.subBand = 2;
        dtuConfig.loraWanPara.eirp    = 22;
        break;
    case LORAMAC_REGION_AU915:
        setSubBand(2);
        dtuConfig.loraWanPara.subBand = 2;
        dtuConfig.loraWanPara.eirp    = 22;
        break;
    case LORAMAC_REGION_EU868:
        dtuConfig.loraWanPara.eirp = 16;
        break;
    case LORAMAC_REGION_IN865:
        dtuConfig.loraWanPara.eirp = 22;
        break;
    case LORAMAC_REGION_RU864:
        dtuConfig.loraWanPara.eirp = 16;
        break;
    default:
        break;
    }

    for (int i = 0; i < 4; i++) {
        dtuConfig.loopOperation.io[i].enable = 0;
    }
    for (int i = 0; i < 4; i++) {
        dtuConfig.loopOperation.ioCount[i].enable = 0;
    }
    for (int i = 0; i < 4; i++) {
        dtuConfig.loopOperation.ioIn[i].enable = 0;
    }

    dtuConfig.hardPara.loopTime            = 60000; // ms
    dtuConfig.hardPara.coding              = 1;
    dtuConfig.hardPara.mcuSleep            = 0;
    dtuConfig.hardPara.baud[0]             = 115200;
    dtuConfig.hardPara.uartConfig[0].value = 3;
    dtuConfig.hardPara.addr                = 0;

    uint8_t sn[9]={0};
    GET_UNIQUE_ID(sn);
    md5_md5(sn,8,str);
    memcpy(sn,str,8);
    // { 
    //     printf("SN/DEVEUI:\n");
    //     for (int i = 0; i < 8; i++) {
    //         printf("%02X",sn[i]);
    //     }
    //     printf("\n");
    // }
    memcpy1(dtuConfig.hardPara.sn,sn,8);
#ifdef SN_AS_EUI
    memcpy1(dtuConfig.loraWanPara.devEui, dtuConfig.hardPara.sn, 8);
#endif
    dtuConfig.loraWanPara.macVersion[0] = 1;
    dtuConfig.loraWanPara.macVersion[1] = 0;
    dtuConfig.loraWanPara.macVersion[2] = 3;
    dtuConfig.loraWanPara.joinType      = ACTIVATION_TYPE_OTAA;
    dtuConfig.loraWanPara.port          = 3;
    dtuConfig.loraWanPara.uartPort[0]   = 22;

    // 是否打印log信息
    dtuConfig.logPrint = 1;

    dtuConfig.loopOperation.reTrans             = 0;
    DBG("dtuConfig.loopOperation.adnDisable=%d\n",dtuConfig.loopOperation.adnDisable);
    dtuConfig.loopOperation.adnDisable          = false;
    DBG("dtuConfig.loopOperation.adnDisable=%d\n",dtuConfig.loopOperation.adnDisable);
    dtuConfig.loopOperation.coord.altitude      = 500;
    dtuConfig.loopOperation.coord.latitude      = 30.67;
    dtuConfig.loopOperation.coord.longitude     = 104.06;
    dtuConfig.loopOperation.power[0].supplyTime = -1;
    dtuConfig.loopOperation.power[1].supplyTime = -1;
    dtuConfig.loopOperation.power[0].enable     = 0;
    dtuConfig.loopOperation.power[1].enable     = 0;
    dtuConfig.loopOperation.rs485Retry          = 1;

    for (uint8_t i = 0; i < 2; i++) {
        dtuConfig.loopOperation.adc[i].max = 5000;
        dtuConfig.loopOperation.adc[i].max = 0;
    }

    deletLogFile();
}

uint16_t loadDtuConfig(void) {
    uint16_t offset = 0;
    uint8_t data[2];
    sDtuConfig_t config;
    NvmmRead((uint8_t *)&config, sizeof(sDtuConfig_t), DF_BOOT_SIZE);

    uint16_t crcTemp = config.crc;
    config.crc       = 0;
    uint16_t crc     = calculateCRC((uint8_t *)&config, sizeof(sDtuConfig_t));
    // sysLog("-----------restore-------------\n");
    DBG("calc_crc=%x  crc=%x\n", crc, crcTemp);
    if (crcTemp == crc) {
        NvmmRead((uint8_t *)&dtuConfig, sizeof(sDtuConfig_t), DF_BOOT_SIZE);
        if (!isUartConfigValid(config.hardPara.baud[0], config.hardPara.uartConfig[0])) {
            dtuConfig.hardPara.baud[0] = 115200;
            // dtuConfig.hardPara.uartConfig =  (8<<4) | (1&0xf);
        }
    } else {
        recoverDtuConfig();
    }

    return 0;
}

uint16_t NvmDataMgmtRestore(void) {
    /*
    #if( CONTEXT_MANAGEMENT_ENABLED == 1 )
        MibRequestConfirm_t mibReq;
        mibReq.Type = MIB_NVM_CTXS;
        LoRaMacMibGetRequestConfirm( &mibReq );
        LoRaMacNvmData_t* nvm = mibReq.Param.Contexts;
        uint16_t offset = 0;
        uint8_t data[2];
        sDtuConfig_t config;
        NvmmRead((uint8_t*)&config,sizeof(sDtuConfig_t),DF_BOOT_SIZE);

        uint16_t crcTemp = config.crc;
        config.crc = 0;
        uint16_t crc = calculateCRC((uint8_t*)&config,sizeof(sDtuConfig_t));
        //sysLog("-----------restore-------------\n");
        DBG("calc_crc=%x  crc=%x\n",crc,crcTemp);
        if(crcTemp == crc){
            NvmmRead( ( uint8_t* ) &dtuConfig, sizeof( sDtuConfig_t),DF_BOOT_SIZE);
            if(!isUartConfigValid(config.hardPara.baud[0],config.hardPara.uartConfig[0])){
                dtuConfig.hardPara.baud[0] = 115200;
                //dtuConfig.hardPara.uartConfig =  (8<<4) | (1&0xf);
            }
            return 1;
        }
       // if( NvmmRead( ( uint8_t* ) nvm, sizeof( LoRaMacNvmData_t ), 0 ) ==
           //           sizeof( LoRaMacNvmData_t ) )
       // {
          //  return sizeof( LoRaMacNvmData_t );
       // }
    #endif
        return 0;
    */
    return 1;
}

bool checkDtuConfigCrc(void) {
    sDtuConfig_t config;
    NvmmRead((uint8_t *)&config, sizeof(sDtuConfig_t), DF_BOOT_SIZE);
    uint16_t crcTemp          = config.crc;
    config.crc                = 0;
    uint16_t crc              = calculateCRC((uint8_t *)&config, sizeof(sDtuConfig_t));
    if (crcTemp == crc) {
        return true;
    } else {
        return false;
    }
}
bool NvmDataMgmtFactoryReset(void) {
    uint16_t offset = 0;
#if (CONTEXT_MANAGEMENT_ENABLED == 1)
    // Crypto
    if (NvmmReset(sizeof(LoRaMacCryptoNvmData_t), offset) == false) {
        return false;
    }
    offset += sizeof(LoRaMacCryptoNvmData_t);

    // Mac Group 1
    if (NvmmReset(sizeof(LoRaMacNvmDataGroup1_t), offset) == false) {
        return false;
    }
    offset += sizeof(LoRaMacNvmDataGroup1_t);

    // Mac Group 2
    if (NvmmReset(sizeof(LoRaMacNvmDataGroup2_t), offset) == false) {
        return false;
    }
    offset += sizeof(LoRaMacNvmDataGroup2_t);

    // Secure element
    if (NvmmReset(sizeof(SecureElementNvmData_t), offset) == false) {
        return false;
    }
    offset += sizeof(SecureElementNvmData_t);

    // Region group 1
    if (NvmmReset(sizeof(RegionNvmDataGroup1_t), offset) == false) {
        return false;
    }
    offset += sizeof(RegionNvmDataGroup1_t);

    // Region group 2
    if (NvmmReset(sizeof(RegionNvmDataGroup2_t), offset) == false) {
        return false;
    }
    offset += sizeof(RegionNvmDataGroup2_t);

    // Class b
    if (NvmmReset(sizeof(LoRaMacClassBNvmData_t), offset) == false) {
        return false;
    }
    offset += sizeof(LoRaMacClassBNvmData_t);
#endif
    return true;
}
