#include "rtc-board.h"
#include "CH58x_common.h"

#include <stddef.h>
#include <stdbool.h>
#include <stdint.h>
#include "CONFIG.h"
#include "HAL.h"


/*!
 * \file      rtc-board.c
 *
 * \brief     Target board RTC timer and low power modes management
 *
 * \copyright Revised BSD License, see section \ref LICENSE.
 *
 * \code
 *                ______                              _
 *               / _____)             _              | |
 *              ( (____  _____ ____ _| |_ _____  ____| |__
 *               \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 *               _____) ) ____| | | || |_| ____( (___| | | |
 *              (______/|_____)_|_|_| \__)_____)\____)_| |_|
 *              (C)2013-2017 Semtech
 *
 * \endcode
 *
 * \author    Miguel Luis ( Semtech )
 *
 * \author    Gregory Cristian ( Semtech )
 *
 * \author    Marten Lootsma(TWTG) on behalf of Microchip/Atmel (c)2017
 */


#include "board.h"
#include "timer.h"
#include "systime.h"
#include "gpio.h"

#include "rtc-board.h"

#define RTC_DEBUG_ENABLE                            1
#define RTC_DEBUG_DISABLE                           0

#define RTC_DEBUG_GPIO_STATE                        RTC_DEBUG_DISABLE
#define RTC_DEBUG_PRINTF_STATE                      RTC_DEBUG_DISABLE

#define MIN_ALARM_DELAY                             3 // in ticks

/*!
 * \brief Indicates if the RTC is already Initialized or not
 */
static bool RtcInitialized = false;
static volatile bool RtcTimeoutPendingInterrupt = false;
static volatile bool RtcTimeoutPendingPolling = false;

typedef enum AlarmStates_e
{
    ALARM_STOPPED = 0,
    ALARM_RUNNING = !ALARM_STOPPED
} AlarmStates_t;

/*!
 * RTC timer context
 */
typedef struct
{
    uint32_t Time;  // Reference time
    uint32_t Delay; // Reference Timeout duration
    uint32_t AlarmState;
}RtcTimerContext_t;

/*!
 * Keep the value of the RTC timer when the RTC alarm is set
 * Set with the \ref RtcSetTimerContext function
 * Value is kept as a Reference to calculate alarm
 */
static RtcTimerContext_t RtcTimerContext;

#if( RTC_DEBUG_GPIO_STATE == RTC_DEBUG_ENABLE )
Gpio_t DbgRtcPin0;
Gpio_t DbgRtcPin1;
#endif

/*!
 * Used to store the Seconds and SubSeconds.
 *
 * WARNING: Temporary fix fix. Should use MCU NVM internal
 *          registers
 */
uint32_t RtcBkupRegisters[] = { 0, 0 };

/*!
 * \brief Callback for the hw_timer when alarm expired
 */
static void RtcAlarmIrq( void );

/*!
 * \brief Callback for the hw_timer when counter overflows
 */
static void RtcOverflowIrq( void );

void RtcInit( void )
{

    //printf("RtcInit\n");
    if( RtcInitialized == false )
    {
        PFIC_DisableIRQ(RTC_IRQn);
        HAL_TimeInit();
        RtcTimerContext.AlarmState = ALARM_STOPPED;
        PFIC_EnableIRQ(RTC_IRQn);
        RtcSetTimerContext( );
        RtcInitialized = true;
        //HAL_SleepInit();
    }
}

uint32_t RtcSetTimerContext( void )
{
    RtcTimerContext.Time = ( uint32_t )RTC_GetCycle32k();
    return ( uint32_t )RtcTimerContext.Time;
}

uint32_t RtcGetTimerContext( void )
{
    return RtcTimerContext.Time;
}

uint32_t RtcGetMinimumTimeout( void )
{
    return( MIN_ALARM_DELAY );
}

uint32_t RtcMs2Tick( TimerTime_t milliseconds )
{
    return ( uint32_t )( ( ( ( uint64_t )milliseconds ) << 15 ) / 1000 );
    //return ( uint32_t )( ( ( ( uint64_t )milliseconds ) << 5 ) );
}

TimerTime_t RtcTick2Ms( uint32_t tick )
{

    uint32_t seconds = tick >> 15;
    //uint32_t seconds = tick >> 5;
    tick = tick & 0x7FFF;
//
    //return seconds;
    return ( ( seconds * 1000 ) + ( ( tick * 1000 ) >> 15 ));
}




void RtcDelayMs( TimerTime_t milliseconds )
{
    uint32_t delayTicks = 0;
    uint32_t refTicks = RtcGetTimerValue( );

    delayTicks = RtcMs2Tick( milliseconds );

    // Wait delay ms
    while( ( ( RtcGetTimerValue( ) - refTicks ) ) < delayTicks )
    {
          __NOP( );
    }
}

void RtcSetAlarm( uint32_t timeout )
{
    RtcStartAlarm( timeout );
}

void RtcStopAlarm( void )
{
    RtcTimerContext.AlarmState = ALARM_STOPPED;
}
#define COMPARE_COUNT_MAX_VALUE ( uint32_t )( -1 )
bool HwTimerLoadAbsoluteTicks(uint32_t ticks)
{
   // RTC_CRITICAL_SECTION_ENTER();
    //printf("ticks = %ld\n",ticks-RTC_GetCycle32k());
    unsigned long irq_status;
   // SYS_DisableAllIrq(&irq_status);
    RTC_TRIGFunCfg(ticks-RTC_GetCycle32k());
    //SYS_RecoverIrq(irq_status);
    uint32_t current = RTC_GetCycle32k();
    //RTC_CRITICAL_SECTION_LEAVE();

    if((ticks - current - 1) >= (COMPARE_COUNT_MAX_VALUE >> 1)) {
        // if difference is more than half of max assume timer has passed
        return false;
    }
    if((ticks - current) < 10) {
        // if too close the matching interrupt does not trigger, so handle same as passed
        return false;
    }
    return true;
}
void RtcStartAlarm( uint32_t timeout )
{
    CRITICAL_SECTION_BEGIN( );
    PFIC_DisableIRQ(RTC_IRQn);
    RtcStopAlarm( );

    RtcTimerContext.Delay = timeout;



    RtcTimeoutPendingInterrupt = true;
    RtcTimeoutPendingPolling = false;

    RtcTimerContext.AlarmState = ALARM_RUNNING;
   // printf("rtc=%ld\n",RtcTimerContext.Time);
    //printf("rtc1=%ld\n",RtcGetTimerValue());
   // printf("de=%ld\n",RtcTimerContext.Delay);
    if( HwTimerLoadAbsoluteTicks( RtcTimerContext.Time + RtcTimerContext.Delay ) == false )
    {
        // If timer already passed
        if( RtcTimeoutPendingInterrupt == true )
        {
           // And interrupt not handled, mark as polling
            RtcTimeoutPendingPolling = true;
            RtcTimeoutPendingInterrupt = false;
        }
    }
    CRITICAL_SECTION_END( );
    PFIC_EnableIRQ(RTC_IRQn);
}

uint32_t RtcGetTimerValue( void )
{
    return ( uint32_t )RTC_GetCycle32k();
}

uint32_t RtcGetTimerElapsedTime( void )
{
    return ( uint32_t)( ( uint32_t )RTC_GetCycle32k()- RtcTimerContext.Time );
}

TimerTime_t getSysTime(void){
    //printf("%d\n",(RtcGetTimerValue()*1000)>>15);

   uint32_t tick = RtcGetTimerValue();
   uint32_t seconds = tick >> 15;
   tick = tick & 0x7FFF;
//
   //return seconds;
   return ( ( seconds * 1000 ) + ( ( tick * 1000 ) >> 15 ));
}


uint32_t RtcGetCalendarTime( uint16_t *milliseconds )
{
    uint32_t ticks = 0;

    uint32_t calendarValue = ( uint32_t )RTC_GetCycle32k();

    uint32_t seconds = ( uint32_t )calendarValue >> 10;

    ticks =  ( uint32_t )calendarValue & 0x3FF;

    *milliseconds = RtcTick2Ms( ticks );

    return seconds;
}

void RtcBkupWrite( uint32_t data0, uint32_t data1 )
{
    CRITICAL_SECTION_BEGIN( );
    RtcBkupRegisters[0] = data0;
    RtcBkupRegisters[1] = data1;
    CRITICAL_SECTION_END( );
}

void RtcBkupRead( uint32_t* data0, uint32_t* data1 )
{
    CRITICAL_SECTION_BEGIN( );
    *data0 = RtcBkupRegisters[0];
    *data1 = RtcBkupRegisters[1];
    CRITICAL_SECTION_END( );
}

void RtcProcess( void )
{
    CRITICAL_SECTION_BEGIN( );
    if( (  RtcTimerContext.AlarmState == ALARM_RUNNING ) && ( RtcTimeoutPendingPolling == true ) )
    {
        PFIC_DisableIRQ(RTC_IRQn);
        if( RtcGetTimerElapsedTime( ) >= RtcTimerContext.Delay )
        {
            RtcTimerContext.AlarmState = ALARM_STOPPED;

            // Because of one shot the task will be removed after the callback
            RtcTimeoutPendingPolling = false;
            // NOTE: The handler should take less then 1 ms otherwise the clock shifts
            TimerIrqHandler( );
        }
        PFIC_EnableIRQ(RTC_IRQn);
    }
    CRITICAL_SECTION_END( );
}

TimerTime_t RtcTempCompensation( TimerTime_t period, float temperature )
{
    return period;
}
uint32_t timer0Count = 0;
__INTERRUPT
__HIGH_CODE
void RTC_IRQHandler(void)
{
    R8_RTC_FLAG_CTRL = (RB_RTC_TMR_CLR | RB_RTC_TRIG_CLR);



     //timer0Count++;
    RtcTimerContext.AlarmState = ALARM_STOPPED;
    // Because of one shot the task will be removed after the callback
    RtcTimeoutPendingInterrupt = false;

    // NOTE: The handler should take less then 1 ms otherwise the clock shifts
    TimerIrqHandler( );

}

static void RtcOverflowIrq( void )
{
    //RtcTimerContext.Time += ( uint64_t )( 1 << 32 );
}
